/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/RCS/agw-main.cc,v 1.33 1996/05/23 01:48:02 elan Exp $ (LBL)";

#include <osfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/stat.h>
#ifdef sgi
#include <getopt.h>
#endif

#if defined(sun) && !defined(__svr4__)
extern "C" int gettimeofday(struct timeval*, struct timezone*);
extern "C" int nice(int incr);
#endif
#if defined(sun) || defined(__osf__)
extern "C" int gethostname(char* name, int len);
#endif

#include "inet.h"
#include "rtp.h"
#ifdef INFOPAD
#include "ipadchan.h"
#endif
#ifdef notyet
#include "crypt.h"
#endif

#include "Tcl.h"
#include "timer.h"

extern "C" int eventLoop();
extern "C" char version[];

static const char* prog = "agw";

static void
usage()
{
	fprintf(stderr, "\
Usage: agw [-A session] [-I channel] [-g gain] [-u script] [-v] \n");
	exit(1);
}

SIGRET
adios(...)
{
	exit(0);
}

/* Very simple "after" command just to get "alive" beacon working. */
static class AfterCommand : public TclObject, public Timer {
public:
	AfterCommand() : TclObject("after") {}
	int command(int argc, const char*const* argv) {
		if (argc != 3)
			return (TCL_ERROR);
		strcpy(buf_, argv[2]);
		msched(atoi(argv[1]));
		return (TCL_OK);
	}
	void timeout(void) {
		Tcl& t = Tcl::instance();
		t.evalc(buf_);
	}
private:
	char buf_[256];
} cmd_after;

static class AdiosCommand : public TclObject {
public:
	AdiosCommand() : TclObject("adios") {}
	int command(int argc, const char*const* argv) {
		adios();
		/*NOTREACHED*/
		return (0);
	}
} cmd_adios;
 
static class VersionCommand : public TclObject {
public:
	VersionCommand() : TclObject("version") {}
	int command(int argc, const char*const* argv) {
		Tcl::instance().result(version);
		return (TCL_OK);
	}
} cmd_version;
 

static class OptionCommand : public TclObject {
public:
	OptionCommand() : TclObject("option") {}
	int command(int argc, const char*const* argv) {
		Tcl& tcl = Tcl::instance();
		if (argc == 3) {
			if (strcmp(argv[1], "get") != 0)
				return (TCL_ERROR);
			tcl.result(tcl.attr(argv[2]));
			return (TCL_OK);
		} else if (argc == 4) {
			if (strcmp(argv[1], "add") != 0)
				return (TCL_ERROR);
			tcl.add_option(argv[2], argv[3]);
			return (TCL_OK);
		}
		return (TCL_ERROR);
	}
} cmd_option;


extern "C" char *optarg;
extern "C" int optind;
extern "C" int opterr;
extern "C" char tcl_rtpgw[];

int
main(int argc, const char** argv)
{
	const char* script = 0;

	signal(SIGHUP, adios);
	signal(SIGINT, adios);
	signal(SIGTERM, adios);
	signal(SIGQUIT, adios);

	opterr = 0;
	const char* optstr = "A:I:g:u:v";
	Tcl::init("agw");
	Tcl& tcl = Tcl::instance();

	tcl.add_default("confBusChannel", "3");
	tcl.add_default("outputNetworkType", "ip");
	tcl.add_default("filterGain", "0.25");
	tcl.add_default("sessionType", "rtp");
	int op;
	while ((op = getopt(argc, (char**)argv, (char*)optstr)) != -1) {
		switch (op) {
		default:
			usage();
		case 'A':
			tcl.add_option("sessionType", optarg);
			break;
		case 'I':
			tcl.add_option("confBusChannel", optarg);
			break;
		case 'g':
			tcl.add_option("filterGain", optarg);
			break;
		case 'K':
			/*XXX probably do not want this in X server*/
			tcl.add_option("sessionKey", optarg);
			break;
		case 'm':
			tcl.add_option("mtu", optarg);
			break;
		case 'o':
			tcl.add_option("outfile", optarg);
			break;
		case 'u':
			script = optarg;
			break;
		case 'v':
			printf("Agw version %s\n", version);
			exit(0);
			break;
		}
	}

	if (argc - optind != 0)
	    usage();

	EmbeddedTcl::init();
	tcl.evalc(tcl_rtpgw);

	if (script != 0)
		tcl.EvalFile(script);

	tcl.evalf("appInit agw audio %d", atoi(tcl.attr("confBusChannel")));

	for (;;) {
		Tcl_DoOneEvent(TCL_FILE_EVENTS|TCL_TIMER_EVENTS);
	}

	adios();
}
