/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/engines/audio/RCS/transcoder-adpcm.cc,v 1.2 1996/02/27 03:33:10 elan Exp $ (LBL)";
#endif

#include "inet.h"
#include "transcoder.h"
#include "transcoder-pcm.h"
#include "transmitter.h"
#include "encoder.h"
extern "C" {
#include "adpcm.h"
}

class ADPCMTranscoder : public PCMTranscoder {
public:
	ADPCMTranscoder();
	virtual int recv(const rtphdr*, const u_char* pkt, int len, 
			 Transmitter& tx);
protected:
	adpcm_state rstate_;

};

static class ADPCMTranscoderMatcher : public Matcher {
public:
	ADPCMTranscoderMatcher() : Matcher("transcoder") {}
	TclObject* match(const char* fmt) {
		if (strcmp("adpcm", fmt) == 0) {
			return (new ADPCMTranscoder);
		}
		return (0);
	}
} adpcm_transcoder_matcher;

ADPCMTranscoder::ADPCMTranscoder()
{
	memset(&rstate_, 0, sizeof(rstate_));
}


int ADPCMTranscoder::recv(const struct rtphdr* rh, const u_int8_t* frame, 
			     int cc, Transmitter& tx)
{
	u_char wrkbuf[FRAMESIZE*8];

	if ((cc = (cc - sizeof(struct adpcm_state)) << 1) <= 0 ||
	    cc > sizeof(wrkbuf))
		return (0);

	if (tx.bps() == 0)
		return (1);

	double now = gettimeofday();

	if (fc_ <= now) {
		/* If we have fallen behind (>200ms), re-sync. */
		if (now - fc_ > 200000.) 	
			fc_ = now;

		NTOHS(((struct adpcm_state*)frame)->valprev);
		adpcm_decoder(frame + sizeof(struct adpcm_state),
			      wrkbuf, cc, (struct adpcm_state*)frame);

		AudioFrame af(ntohl(rh->rh_ts), wrkbuf, cc);
		int ecc = encoder_->consume(&af, tx);
		fc_ += 1e6 * 8. * ecc / tx.bps();
	}
	return (1);
}


