/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/engines/audio/RCS/transcoder-pcm.cc,v 1.1 1996/02/27 03:32:21 elan Exp $ (UCB)";
#endif

#include "Tcl.h"
#include "transcoder-pcm.h"
#include "transmitter.h"
#include "encoder.h"
#include "config.h"

static class PCMTranscoderMatcher : public Matcher {
    public:
	PCMTranscoderMatcher() : Matcher("transcoder") {}
	TclObject* match(const char* fmt) {
		if (strcmp(fmt, "pcm") == 0) {
			return (new PCMTranscoder);
		}
		return (0);
	}
} pcm_transcoder_matcher;

int PCMTranscoder::recv(const rtphdr* rh, const u_int8_t* data, int cc, 
			Transmitter& tx)
{
	if (tx.bps() == 0)
		return (1);

	double now = gettimeofday();

	if (fc_ <= now) {
		/* If we have fallen behind (>200ms), re-sync. */
		if (now - fc_ > 200000.) 	
			fc_ = now;

		AudioFrame af(ntohl(rh->rh_ts), data, cc);
		int ecc = encoder_->consume(&af, tx);
		fc_ += 1e6 * 8. * ecc / tx.bps();
	}
	return (1);
}
