/*
 * Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
#ifndef lint
static char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/vgw/ipad/RCS/encoder-ipadvq.cc,v 1.12 1995/11/22 03:28:25 elan Exp $ (UC Berkeley)";
#endif
/*
 * This file contains C code that implements
 * the VQ compression portion of mpeg2vq.
 * Most of the code here was taken from rgb2vec.
 * Fixed up much by Jeff Gilbert 3/29/95
 * Many format adjustments Jeff Gilbert 4/8/95
 * Adaptation Jeff Gilbert 4/27/95
 */
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "net.h"
#include "encoder-ipadvq.h"
#include "vq_file_fmt.h"
#include "ipadchan.h"

#include "global.h"
#include "packets.h"

#ifdef notdef
extern int totNumFrames;	/* This is number of frames that have been */
#endif

/* XXX */
int quietFlag = 1;

#ifdef notdef
 /* Maps MPEG picture rate to real number */
double  picture_rate_table[] = {23.976, 24.0, 25.0, 29.97, 30.0, 50.0, 59.94,
60.0};
#endif
#ifndef min
#define min(a,b)  (((a)<(b)) ? (a) : (b))
#endif
#ifndef max
#define max(a,b)  (((a)>(b)) ? (a) : (b))
#endif


IpadVQEncoder::IpadVQEncoder(IpadChannel* n, int w, int h, int dec, int dump) 
	      :Encoder((Network*)n), 
	      srcX(0), srcY(0), srcWidth(0), srcHeight(0), 
	      outWidth(128), outHeight(240), skipped_frames(0), 
	      dump_codebook(1), halfres_coding(1), uniform_coding(0), 
	      fast_coding(1), no_VQconv(0), frames_per_sec_override(0), 
	      adapt_frames(0), error_tolerance(0), print_vq_error(0), 
	      adapt_global_threshold(0), skip_every(0), net_(n), dump_(dump), 
	      decimation_(dec)
{
#ifdef STATISTICS
	low_error = 0x0fffffff;
	max_error = 0;
	total_error = 0;
	total_vectors = 0;
#endif
	squares = &_squares[256];

	/* XXX */
	if (getenv("ADAPTIVE_CODING") != 0) {
		adapt_frames=2;
		adapt_global_threshold = 1;
		char* tmp_str;
		if ((tmp_str=getenv("ADAPTION_THRESHOLD")) != 0)
		    sscanf(tmp_str,"%d",&adapt_global_threshold);
		halfres_coding=0;
		fprintf(stderr,"Using Adaptive VQ coding threshold %d\n",
			adapt_global_threshold);

		if (getenv("ADAPTION_PRINT_VQ_ERROR"))
		    print_vq_error = 1;
	}

	/* If we only needed to create the encoder, return now. */
	if (w != 0 && h != 0) 
		InitVQ(w, h);
}

IpadVQEncoder::~IpadVQEncoder()
{
	delete L_Y_scaled_buf;
	delete Cr_I_scaled_buf;
	delete Cb_Q_scaled_buf;
	delete Y_frame_data_codes;
	delete I_frame_data_codes;
	delete Q_frame_data_codes;
	delete IQ_temp_codes;
	delete Y_frame_coding_errors;
	delete I_frame_coding_errors;
	delete Q_frame_coding_errors;
	delete L_rescale_info;
	delete CrCb_rescale_info;
}

RescaleInfo::~RescaleInfo()
{
	delete linebuf;
	delete x_orig2scaledlinebuf;
	delete x_scaled2origoffset;
	delete x_scaled2val_scale_large;
	delete x_scaled2val_scale_small;
	delete orig_rows_per_scaled_y;
	delete shift_bits_large_y;
	delete shift_bits_small_y;
}

/*
 *--------------------------------------------------------------
 *
 * InitVQ
 *
 *      Initializes the global variables y_codes, i_codes, and
 *      q_codes with data from the code book.  For now the code
 *      book data must reside in the file "Codebook".
 *      Also initializes the global variable squares.  It also
 *      mallocs some space for vector_buffer to point to.
 *
 * Results:
 *      The aforementioned variables are initialized.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
void
IpadVQEncoder::InitVQ(int width, int height)
{
	FILE   *fp;
	int     i, j;
	int     pixels_per_block;
	VQ_file_header vq_file_header;
#ifdef notdef
	if (no_VQconv) {
		return;
	}
#endif
	if (adapt_frames && halfres_coding) {
		fprintf(stderr,
			"Codebook adaptation does not support halfres.  Using full res\n");
		halfres_coding = 0;
	}
	if (print_vq_error && (adapt_frames == 0) &&
	    (halfres_coding || fast_coding || uniform_coding)) {
		fprintf(stderr,
			"-print_vq_error not supported under fast, halfres, or uniform coding\n");
		print_vq_error = 0;
	}
	/* Configure scaling variables. */
	inWidth = width;
	inHeight = height;
	if ((srcWidth == 0) || (srcHeight == 0)) {
		srcWidth = inWidth;
		srcHeight = inHeight;
	}
	/* Initialize codebook. */
	if (uniform_coding) {
		for (i = 0; i < 256; i++) {
			for (j = 0; j < 16; j++) {
				y_codes[i << 4 | j] = i;
				i_codes[i << 4 | j] = i;
				q_codes[i << 4 | j] = i;
			}
		}
		/* Uniform adaptive coding means only to use uniform codebook
		 * and */
		/* adapt off of that and then slow code */
		if (adapt_frames) {
			uniform_coding = 0;
			fast_coding = 0;
		}
	} else if (fast_coding) {
		generate_fast_codebooks(y_codes, i_codes, q_codes);
		/* Fast adaptive coding means only to use fast codebook and */
		/* adapt off of that and then slow code */
		if (adapt_frames) {
			uniform_coding = 0;
			fast_coding = 0;
		}
	} else {
		if (!(fp = fopen(codebook_file, "r"))) {
			fprintf(stderr, "Cannot open `%s'.\n", codebook_file);
			exit(-1);
		}
		if (fread(y_codes, sizeof(y_codes), 1, fp) < 1 ||
		    fread(i_codes, sizeof(i_codes), 1, fp) < 1 ||
		    fread(q_codes, sizeof(q_codes), 1, fp) < 1) {
			fprintf(stderr, "Unable to read `%s'.\n", codebook_file);
			exit(-1);
		}
		fclose(fp);
	}
	/* Write header */
	vq_file_header.magic_number = htonl(VQ_MAGIC_NUM);
	vq_file_header.major_version = htons(1);
	vq_file_header.minor_version = htons(0);
	vq_file_header.width = htonl(outWidth / 8);
	vq_file_header.height = htonl(outHeight / 8);
	vq_file_header.frame_size = htonl((outWidth / 8) * (outHeight / 8) * 6);
	vq_file_header.codebook_entries = htonl(768);
	vq_file_header.codebook_entry_size = htonl(16);
	vq_file_header.extra_data_len = htonl(0);
	if (frames_per_sec_override != 0) {
		vq_file_header.frames_per_sec =
		    htonl((int) (frames_per_sec_override * 65536 / (skip_every + 1)));
#ifdef notdef
	} else if ((picture_rate > 0) &&
		   (picture_rate <= sizeof(picture_rate_table) /
		    sizeof(*picture_rate_table))) {
		vq_file_header.frames_per_sec =
		    htonl((int) (picture_rate_table[picture_rate - 1] * 65536 /
				 (skip_every + 1)));
#endif
	} else {
		vq_file_header.frames_per_sec = 0;
	}
	/* XXX Transmission not accounted for */
	if (dump_)
		fwrite(&vq_file_header, sizeof(vq_file_header), 1, stdout);

	if (dump_codebook && (!adapt_frames)) {
		if (dump_) {
			/* Say there will be update of codebook entries0-767 */
			putw(htonl(768), stdout);
			putw(htonl(0), stdout);
			fwrite(y_codes, sizeof(y_codes), 1, stdout);
			fwrite(i_codes, sizeof(i_codes), 1, stdout);
			fwrite(q_codes, sizeof(q_codes), 1, stdout);
			fflush(stdout);
		}
    		net_->send_codebook_entries(0, 0, 256, y_codes);
    		net_->send_codebook_entries(1, 0, 256, i_codes);
    		net_->send_codebook_entries(2, 0, 256, q_codes);
	}
	/* For half-res coding, reduce all codebooks by 2x2 and put */
	/* in beginning quarter */
	if (halfres_coding || fast_coding) {
		for (i = 0; i < 256; i++) {
			y_codes[i * 4] = ((int) y_codes[i * 16] + (int) y_codes[i * 16 + 1] +
					  (int) y_codes[i * 16 + 4] + (int) y_codes[i * 16 + 5]) >> 2;
			y_codes[i * 4 + 1] = ((int) y_codes[i * 16 + 2] + (int) y_codes[i * 16 + 3] +
					      (int) y_codes[i * 16 + 6] + (int) y_codes[i * 16 + 7]) >> 2;
			y_codes[i * 4 + 2] = ((int) y_codes[i * 16 + 8] + (int) y_codes[i * 16 + 9] +
					      (int) y_codes[i * 16 + 12] + (int) y_codes[i * 16 + 13]) >> 2;
			y_codes[i * 4 + 3] = ((int) y_codes[i * 16 + 10] + (int) y_codes[i * 16 + 11] +
					      (int) y_codes[i * 16 + 14] + (int) y_codes[i * 16 + 15]) >> 2;
			i_codes[i * 4] = ((int) i_codes[i * 16] + (int) i_codes[i * 16 + 1] +
					  (int) i_codes[i * 16 + 4] + (int) i_codes[i * 16 + 5]) >> 2;
			i_codes[i * 4 + 1] = ((int) i_codes[i * 16 + 2] + (int) i_codes[i * 16 + 3] +
					      (int) i_codes[i * 16 + 6] + (int) i_codes[i * 16 + 7]) >> 2;
			i_codes[i * 4 + 2] = ((int) i_codes[i * 16 + 8] + (int) i_codes[i * 16 + 9] +
					      (int) i_codes[i * 16 + 12] + (int) i_codes[i * 16 + 13]) >> 2;
			i_codes[i * 4 + 3] = ((int) i_codes[i * 16 + 10] + (int) i_codes[i * 16 + 11] +
					      (int) i_codes[i * 16 + 14] + (int) i_codes[i * 16 + 15]) >> 2;
			q_codes[i * 4] = ((int) q_codes[i * 16] + (int) q_codes[i * 16 + 1] +
					  (int) q_codes[i * 16 + 4] + (int) q_codes[i * 16 + 5]) >> 2;
			q_codes[i * 4 + 1] = ((int) q_codes[i * 16 + 2] + (int) q_codes[i * 16 + 3] +
					      (int) q_codes[i * 16 + 6] + (int) q_codes[i * 16 + 7]) >> 2;
			q_codes[i * 4 + 2] = ((int) q_codes[i * 16 + 8] + (int) q_codes[i * 16 + 9] +
					      (int) q_codes[i * 16 + 12] + (int) q_codes[i * 16 + 13]) >> 2;
			q_codes[i * 4 + 3] = ((int) q_codes[i * 16 + 10] + (int) q_codes[i * 16 + 11] +
					      (int) q_codes[i * 16 + 14] + (int) q_codes[i * 16 + 15]) >> 2;
		}
	}
	/* Initialize squares. (Note - it goes from -256..255 */
	for (i = -256; i < 256; i++) {
		squares[i] = i * i;
	}
	if (!quietFlag) {
#ifdef notdef
		if ((picture_rate > 0) &&
		    (picture_rate <= sizeof(picture_rate_table) /
		     sizeof(*picture_rate_table))) {
			fprintf(stderr, "MPEG coded at %.3f frames/sec\n",
				picture_rate_table[picture_rate - 1]);
		} else {
			fprintf(stderr, "MPEG frame rate information not valid\n");
		}
#endif
		if (frames_per_sec_override) {
			fprintf(stderr, "Overriding it with %.3f frames/sec\n",
				frames_per_sec_override);
		}
		fprintf(stderr, "MPEG Original is %dx%d.  VQ is %dx%d\n",
			srcWidth, srcHeight, outWidth, outHeight);
	}
	if (halfres_coding) {
		outWidth /= 2;
		outHeight /= 2;
		if (!quietFlag) {
			fprintf(stderr, "Reducing effective VQ to %dx%d\n",
				outWidth, outHeight);
		}
		pixels_per_block = 4;
	} else {
		pixels_per_block = 16;
	}
	outSize = outWidth * outHeight;
	if (halfres_coding) {
		codeWidth = outWidth / 2;
		codeHeight = outHeight / 2;
	} else {
		codeWidth = outWidth / 4;
		codeHeight = outHeight / 4;
	}
	codeSize = codeWidth * codeHeight;
	/* Must round up to next highest multiple of 16 due to MPEG decoder */
	width = ((width + 15) / 16) * 16;
	/* Make temporary buffers for scaling. */
	L_rescale_info = init_rescale_info(srcWidth, srcHeight, width,
					   outWidth, outHeight, outWidth);
	if (L_rescale_info == NULL) {
		fprintf(stderr, "Unable to allocate resize buffers\n");
		exit(2);
	}
	int d = (decimation_ == 411)? 2 : 1;
	CrCb_rescale_info = init_rescale_info(srcWidth / 2, srcHeight / d, 
					      width / 2, outWidth / 2, 
					      outHeight / 2, outWidth / 2);
	if (CrCb_rescale_info == NULL) {
		fprintf(stderr, "Unable to allocate resize buffers\n");
		exit(2);
	}
	L_Y_scaled_buf = new u_char[outSize * max(1, adapt_frames)];
	Cr_I_scaled_buf = new u_char[outSize / 4 * max(1, adapt_frames)];
	Cb_Q_scaled_buf = new u_char[outSize / 4 * max(1, adapt_frames)];
	memset(L_Y_scaled_buf, 128, outSize);
	memset(Cr_I_scaled_buf, 128, outSize / 4);
	memset(Cb_Q_scaled_buf, 128, outSize / 4);
	Y_frame_data_codes = new u_char[codeSize * max(1, adapt_frames)];
	I_frame_data_codes = new u_char[codeSize / 4 * max(1, adapt_frames)];
	Q_frame_data_codes = new u_char[codeSize / 4 * max(1, adapt_frames)];
	IQ_temp_codes = new u_char[codeSize / 2];
	if ((L_Y_scaled_buf == NULL) ||
	    (Cr_I_scaled_buf == NULL) ||
	    (Cb_Q_scaled_buf == NULL) ||
	    (Y_frame_data_codes == NULL) ||
	    (I_frame_data_codes == NULL) ||
	    (Q_frame_data_codes == NULL) ||
	    (IQ_temp_codes == NULL)) {
		fprintf(stderr, "Unable to allocate resized image buffers\n");
		exit(2);
	}
	if (adapt_frames) {
		Y_frame_coding_errors = new int[codeSize * adapt_frames];
		I_frame_coding_errors = new int [codeSize / 4 * adapt_frames];
		Q_frame_coding_errors = new int[codeSize / 4 * adapt_frames];
	} else {
		Y_frame_coding_errors =
		    I_frame_coding_errors =
		    Q_frame_coding_errors = NULL;
	}
	/* Initialize CrCb to IQ table */
	init_CrCb_to_IQ();
	prev_codes_valid = 0;
	adapt_frames_stored = 0;
	adapt_frames_needed = adapt_frames;
}
/*
 *--------------------------------------------------------------
 *
 * init_CrCb_to_IQ
 *
 *      Called to initialize image rescale aux buffers
 *
 * Results:
 *      CrCb_to_IQ is initialized.
 *
 * Math behind transformations from MPEG (L, Cr, Cb) space to Infopad (Y,I,Q)
 * space (ala JMG):
 *      MPEG Defines:   R = L + 1.402*(Cb-128)
 *                      G = L - 0.344*(Cr-128) - 0.714*(Cb-128)
 *                      B = L + 1.772*(Cr-128)
 *
 *      InfoPad Says:   R = Y + 1.188*(I-128) + 0.719*(Q-128)
 *                      G = Y - 0.328*(I-128) - 0.750*(Q-128)
 *                      B = Y - 1.375*(I-128) + 1.969*(Q-128)
 *
 *      Temporarily disregarding the 128's we get:
 *              | R |   | 1.000   0.000   1.402 |   | L  |
 *              | G | = | 1.000  -0.344  -0.714 | * | Cr |
 *              | B |   | 1.000   1.772   0.000 |   | Cb |
 *
 *                              And
 *
 *              | R |   | 1.000   1.188   0.719 |   | Y |
 *              | G | = | 1.000  -0.328  -0.750 | * | I |
 *              | B |   | 1.000  -1.375   1.969 |   | Q |
 *
 *                              So:
 *
 *      | 1.000   0.000   1.402 |   | L  |   | 1.000   1.188   0.719 |   | Y |
 *      | 1.000  -0.344  -0.714 | * | Cr | = | 1.000  -0.328  -0.750 | * | I |
 *      | 1.000   1.772   0.000 |   | Cb |   | 1.000  -1.375   1.969 |   | Q |
 *
 *
 *                                       -1
 *      | Y |   | 1.000   1.188   0.719 |     | 1.000   0.000   1.402 |   | L  |
 *      | I | = | 1.000  -0.328  -0.750 |   * | 1.000  -0.344  -0.714 | * | Cr |
 *      | Q |   | 1.000  -1.375   1.969 |     | 1.000   1.772   0.000 |   | Cb |
 *
 *              | 1.000   0.003  -0.004 |          | 1     0       0   |
 *            = | 0.000  -0.384   0.831 | =approx= | 0  -0.384   0.831 |
 *              | 0.000   0.630   0.583 |          | 0   0.630   0.583 |
 *
 *  So Y = L
 *     I = -0.384*Cr + 0.831*Cb
 *     Q =  0.630*Cr + 0.583*Cb
 *
 * But I,Q,Cr, Cb all have 128 offsets so we have to account:
 *    (I-128) = -0.384*(Cr-128) + 0.831*(Cb-128)
 *    (Q-128) =  0.630*(Cr-128) + 0.583*(Cb-128)
 *
 *                      so
 *
 *    Y = L
 *    I = -0.384*Cr + 0.831*Cb + 70.784
 *    Q =  0.630*Cr + 0.583*Cb - 27.264
 *--------------------------------------------------------------
 */
void 
IpadVQEncoder::init_CrCb_to_IQ(void)
{
	int     Cr, Cb;
	float   i_float, q_float;
	unsigned char i_uchar, q_uchar;
	/* Initialize CrCb to IQ buffer */
	for (Cr = 0; Cr < 256; Cr++) {
		for (Cb = 0; Cb < 256; Cb++) {
#ifdef notdef
			i_float = -0.384 * Cr + 0.831 * Cb + 70.784;
			q_float = 0.630 * Cr + 0.583 * Cb - 27.264;
#endif
			/* YUV -> YIQ */
			i_float = -0.2676 * (Cr - 128) + 0.7361 * (Cb - 128);
			q_float = 0.3869 * (Cr - 128) + 0.4596 * (Cb - 128);
			i_float += 128.;
			q_float += 128.;
			if (i_float < 0) {
				i_uchar = 0;
			} else if (i_float > 255.0) {
				i_uchar = 255;
			} else {
				i_uchar = (unsigned char) i_float;
			}
			if (q_float < 0) {
				q_uchar = 0;
			} else if (q_float > 255.0) {
				q_uchar = 255;
			} else {
				q_uchar = (unsigned char) q_float;
			}
			CrCb_to_IQ[Cr | (Cb << 8)] = (unsigned short) i_uchar |
			    (((unsigned short) q_uchar) << 8);
		}
	}
}
 /* log2_int just computes log to the base 2 of an integer */
int 
IpadVQEncoder::log2_int(u_int n)
{
	int     logval;
	for (logval = -1; n > 0; logval++, n >>= 1);
	return logval;
}
void 
IpadVQEncoder::ffree(void *ptr)
{
	if (ptr != NULL) {
		delete ptr;
	}
}
/*
 *--------------------------------------------------------------
 *
 * init_rescale_info
 *
 *      Called to initialize image rescale aux buffers
 *
 * Results:
 *      Buffers are initted.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
RescaleInfo *
IpadVQEncoder::init_rescale_info(int src_width, int src_height, 
				 int src_bytes_per_line, int dst_width, 
				 int dst_height, int dst_bytes_per_line)
{
	RescaleInfo *ri;
	int     x, y, i, j;
	int     x_ratio_small;
	int     x_ratio_large;
	int     y_ratio_small;
	int     y_ratio_large;
	int     shift_bits_XY, shift_bits_xY, shift_bits_Xy, shift_bits_xy;
	if (!(ri = new RescaleInfo)) {
		return NULL;
	}
	ri->orig_width = src_width;
	ri->orig_height = src_height;
	ri->orig_bytes_per_line = src_bytes_per_line;
	ri->scaled_width = dst_width;
	ri->scaled_height = dst_height;
	ri->scaled_bytes_per_line = dst_bytes_per_line;
	ri->x_orig2scaledlinebuf = NULL;
	ri->x_scaled2origoffset = NULL;
	ri->linebuf = NULL;
	ri->x_scaled2val_scale_large = NULL;
	ri->x_scaled2val_scale_small = NULL;
	ri->orig_rows_per_scaled_y = NULL;
	ri->shift_bits_large_y = NULL;
	ri->shift_bits_small_y = NULL;
	if (!(ri->linebuf = new int[dst_width]) ||
	    !(ri->x_orig2scaledlinebuf = new int*[src_width]) ||
	    !(ri->x_scaled2origoffset = new int[dst_width]) ||
	    !(ri->x_scaled2val_scale_large = new u_char*[dst_width]) ||
	    !(ri->x_scaled2val_scale_small = new u_char*[dst_width]) ||
	    !(ri->orig_rows_per_scaled_y = new int[dst_height]) ||
	    !(ri->shift_bits_large_y = new int[dst_width]) ||
	    !(ri->shift_bits_small_y = new int[dst_width])) {
		ffree(ri->linebuf);
		ffree(ri->x_orig2scaledlinebuf);
		ffree(ri->x_scaled2origoffset);
		ffree(ri->x_scaled2val_scale_large);
		ffree(ri->x_scaled2val_scale_small);
		ffree(ri->orig_rows_per_scaled_y);
		ffree(ri->shift_bits_large_y);
		ffree(ri->shift_bits_small_y);
		ffree(ri);
		return NULL;
	}
	/* x_ratio_small, x_ratio_large, y_ratio_small, and y_ratio_large */
	/* are the number of source pixels which will be added into one */
	/* destination pixel depending if you are on a small or large step */
	if (src_width > dst_width) {
		x_ratio_small = src_width / dst_width;
		x_ratio_large = x_ratio_small + 1;
	} else {
		x_ratio_small = x_ratio_large = 1;
	}
	if (src_height > dst_height) {
		y_ratio_small = src_height / dst_height;
		y_ratio_large = y_ratio_small + 1;
	} else {
		y_ratio_small = y_ratio_large = 1;
	}
	ri->small_rows_per_scaled_y = y_ratio_small;
	shift_bits_xy = log2_int(x_ratio_small * y_ratio_small);
	shift_bits_xY = log2_int(x_ratio_small * y_ratio_large);
	shift_bits_Xy = log2_int(x_ratio_large * y_ratio_small);
	shift_bits_XY = log2_int(x_ratio_large * y_ratio_large);
	for (i = 0; i < 512; i++) {
		ri->xval_scale[i] =
		    min((i << shift_bits_xy) / (x_ratio_small * y_ratio_small), 255);
		ri->xval_scale[i + 512] =
		    min((i << shift_bits_Xy) / (x_ratio_large * y_ratio_small), 255);
		ri->xval_scale[i + 1024] =
		    min((i << shift_bits_xY) / (x_ratio_small * y_ratio_large), 255);
		ri->xval_scale[i + 1536] =
		    min((i << shift_bits_XY) / (x_ratio_large * y_ratio_large), 255);
	}
	if (src_width > dst_width) {
		/* x_scaled2origoffset not used */
		delete ri->x_scaled2origoffset;
		ri->x_scaled2origoffset = NULL;
		/* Compute x_orig2scaledlinebuf and also use line_buf to
		 * figure */
		/* out how wide each pixel is */
		memset(ri->linebuf, 0, sizeof(int) * dst_width);
		for (x = 0; x < src_width; x++) {
			i = x * dst_width / src_width;
			ri->x_orig2scaledlinebuf[x] = ri->linebuf + i;
			ri->linebuf[i]++;
		}
		/* Some will be x_ratio_large and some will be x_ratio_small */
		/* for each pixel, map x_scaled2val_scale_large and */
		/* x_scaled2val_scale_large to the correct xval_scale part */
		for (x = 0; x < dst_width; x++) {
			if (ri->linebuf[x] == x_ratio_small) {
				ri->x_scaled2val_scale_small[x] = ri->xval_scale;
				ri->x_scaled2val_scale_large[x] = ri->xval_scale + 1024;
				ri->shift_bits_small_y[x] = shift_bits_xy;
				ri->shift_bits_large_y[x] = shift_bits_xY;
			} else {
				ri->x_scaled2val_scale_small[x] = ri->xval_scale + 512;
				ri->x_scaled2val_scale_large[x] = ri->xval_scale + 1536;
				ri->shift_bits_small_y[x] = shift_bits_Xy;
				ri->shift_bits_large_y[x] = shift_bits_XY;
			}
		}
	} else {
		/* x_orig2scaledlinebuf not used */
		delete ri->x_orig2scaledlinebuf;
		ri->x_orig2scaledlinebuf = NULL;
		for (x = 0; x < dst_width; x++) {
			ri->x_scaled2origoffset[x] = x * src_width / dst_width;
		}
		for (x = 0; x < dst_width; x++) {
			ri->x_scaled2val_scale_small[x] = ri->xval_scale;
			ri->x_scaled2val_scale_large[x] = ri->xval_scale + 1024;
			ri->shift_bits_small_y[x] = shift_bits_xy;
			ri->shift_bits_large_y[x] = shift_bits_xY;
		}
	}
	/* Compute orig_rows_per_scaled_y. */
	memset(ri->orig_rows_per_scaled_y, 0, sizeof(int) * dst_height);
	for (y = 0; y < src_height; y++) {
		ri->orig_rows_per_scaled_y[y * dst_height / src_height]++;
	}
	return ri;
}


/*
 *--------------------------------------------------------------
 *
 * rescale_image
 *
 *      Called to rescale an image from one size to another using
 * 		precomputed image rescale aux buffers
 *
 * Results:
 *      Destination image is computed.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
void 
IpadVQEncoder::rescale_image(u_char* src_image, u_char* dst_image, 
			     RescaleInfo* ri, int src_x_offset, 
			     int src_y_offset, int dst_x_offset, 
			     int dst_y_offset)
{
	int     x, y;
	u_char *src_ptr;
	u_char *dst_ptr;
	int     src_width = ri->orig_width;
	int     dst_width = ri->scaled_width;
	int    *orig_rows_per_scaled_y_ptr;
	int     src_row_in_dst_row;
	int     orig_is_wider;
	int   **x_orig2scaledlinebuf_ptr;
	int    *x_scaled2origoffset_ptr;
	int    *line_buf_ptr;
	int    *shift_bits_ptr;
	u_char **x_scaled2val_ptr;
	src_ptr = src_image + src_x_offset + src_y_offset * ri->orig_bytes_per_line;
	dst_ptr = dst_image + dst_x_offset + dst_y_offset * ri->scaled_bytes_per_line;
	orig_is_wider = (src_width > dst_width);
	orig_rows_per_scaled_y_ptr = ri->orig_rows_per_scaled_y;
	for (y = ri->scaled_height; y; y--) {
		memset(ri->linebuf, 0, sizeof(int) * dst_width);
		for (src_row_in_dst_row = *orig_rows_per_scaled_y_ptr;
		     src_row_in_dst_row;
		     src_row_in_dst_row--) {
			if (orig_is_wider) {
				x_orig2scaledlinebuf_ptr = ri->x_orig2scaledlinebuf;
				for (x = src_width; x; x--) {
					*(*x_orig2scaledlinebuf_ptr++) += *src_ptr++;
				}
				src_ptr += ri->orig_bytes_per_line - src_width;
			} else {
				line_buf_ptr = ri->linebuf;
				x_scaled2origoffset_ptr = ri->x_scaled2origoffset;
				for (x = dst_width; x; x--) {
					*line_buf_ptr++ += src_ptr[*x_scaled2origoffset_ptr++];
				}
				src_ptr += ri->orig_bytes_per_line;
			}
		}
		if (*orig_rows_per_scaled_y_ptr == ri->small_rows_per_scaled_y) {
			shift_bits_ptr = ri->shift_bits_small_y;
			x_scaled2val_ptr = ri->x_scaled2val_scale_small;
		} else {
			shift_bits_ptr = ri->shift_bits_large_y;
			x_scaled2val_ptr = ri->x_scaled2val_scale_large;
		}
		line_buf_ptr = ri->linebuf;
		for (x = dst_width; x; x--) {
			*dst_ptr++ = (*x_scaled2val_ptr++)[(*line_buf_ptr++) >>
						       (*shift_bits_ptr++)];
		}
		dst_ptr += ri->scaled_bytes_per_line - dst_width;
		/* If this row is repeated then just copy it */
		while ((y > 1) && (*(++orig_rows_per_scaled_y_ptr) == 0)) {
			memcpy(dst_ptr, dst_ptr - ri->scaled_bytes_per_line, dst_width);
			dst_ptr += ri->scaled_bytes_per_line;
			y--;
		}
	}
}
/*
 *--------------------------------------------------------------
 *
 * DoOutputVQ
 *
 *      Called when image is ready to be converted into VQ.
 *
 * Results:
 *      A VQ version of the input is sent out on stdout.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

int
IpadVQEncoder::encode(u_char* frm, int w, int h, u_char* crvec)
{
	int     y, i;
	u_char *tmp_y_ptr, *tmp_i_ptr, *tmp_q_ptr, *tmp_iq_ptr;
	u_char  t0, t1;
	int     stored_frame_num;
	int 	y_err, i_err, q_err;
	static int skip_count = 0;
	int tot = 0;

	if (skip_every) {
		if (skip_count > 0) {
			skip_count++;
			if (skip_count > skip_every) {
				skip_count = 0;
			}
			skipped_frames++;
			return (0);
		}
		skip_count++;
	}
#ifdef notdef
	totNumFrames++;
#endif
	if (no_VQconv) {
		return (0);
	}
	
	int d = (decimation_ == 411)? 4 : 2;
	int size = w * h;
	u_char* l = frm;
	u_char* Cr = l + size;
	u_char* Cb = Cr + size / d;

	rescale_image(l, L_Y_scaled_buf + outSize * adapt_frames_stored,
		      L_rescale_info, srcX, srcY, 0, 0);
	rescale_image(Cr, Cr_I_scaled_buf + outSize / 4 * adapt_frames_stored,
		      CrCb_rescale_info, srcX / 2, srcY / 2, 0, 0);
	rescale_image(Cb, Cb_Q_scaled_buf + outSize / 4 * adapt_frames_stored,
		      CrCb_rescale_info, srcX / 2, srcY / 2, 0, 0);

	convert_CrCb_IQ(Cr_I_scaled_buf + outSize / 4 * adapt_frames_stored,
			Cb_Q_scaled_buf + outSize / 4 * adapt_frames_stored,
			CrCb_rescale_info->scaled_height,
			CrCb_rescale_info->scaled_width,
			CrCb_rescale_info->scaled_bytes_per_line);

	if (adapt_frames_needed) {
		adapt_frames_needed--;
		adapt_frames_stored++;
	}
	/* Need to store up some more */
	if (adapt_frames_needed) {
		return (0);
	}
	if (adapt_frames_stored) {
		if (!quietFlag) {
			fprintf(stderr, "Time to ADAPT!\n");
		}
		adapt_codebook();
		if (!quietFlag) {
			fprintf(stderr, "ADAPTED!\n");
		}
		if (dump_) {
		      /* Say there will be update of codebook entries 0..767 */
			putw(htonl(768), stdout);
			putw(htonl(0), stdout);
			tot += sizeof(long) * 2;
			fwrite(y_codes, sizeof(y_codes), 1, stdout);
			fwrite(i_codes, sizeof(i_codes), 1, stdout);
			fwrite(q_codes, sizeof(q_codes), 1, stdout);
			fflush(stdout);
		}
    		net_->send_codebook_entries(0, 0, 256, y_codes);
    		net_->send_codebook_entries(1, 0, 256, i_codes);
    		net_->send_codebook_entries(2, 0, 256, q_codes);
		tot += sizeof(y_codes) + sizeof(i_codes) + sizeof(q_codes);
	}
	for (stored_frame_num = 0;
	     stored_frame_num < max(1, adapt_frames_stored);
	     stored_frame_num++) {
		y_err = vector_quantize_block(L_Y_scaled_buf + 
			      stored_frame_num * outSize, Y_frame_data_codes, 
			      y_codes, codeWidth, codeHeight, outWidth);
		i_err = vector_quantize_block(Cr_I_scaled_buf + 
			      stored_frame_num * outSize / 4,
			      I_frame_data_codes, 
			      i_codes, codeWidth / 2, codeHeight / 2, 
			      outWidth / 2);
		q_err = vector_quantize_block(Cb_Q_scaled_buf + 
			      stored_frame_num * outSize / 4,
			      Q_frame_data_codes, q_codes, codeWidth / 2, 
			      codeHeight / 2, outWidth / 2);
		if (print_vq_error) {
			fprintf(stderr, 
				"Frame coding errors: Y=%d I=%d Q=%d\n",
				 y_err, i_err, q_err);
		}
	if (adapt_global_threshold && adapt_frames &&
		    (adapt_frames_stored == 0) &&
		    ((y_err > adapt_global_threshold) ||
		     (i_err > adapt_global_threshold) ||
		     (q_err > adapt_global_threshold))) {
			if (!quietFlag) {
				fprintf(stderr, "Forcing adapt due to global error!\n");
			}
			adapt_frames_stored = 1;
			adapt_frames_needed = adapt_frames - 1;
			return (0);
		}
		prev_codes_valid = 1;

		/* 
		 * Write whole frame of Y then whole frame of IQ interleaved
		 * as IIQQ.  Preceede with a long word 0 indicating no (more)
		 */

		/* codebook update */
		if (dump_) 
			putw(htonl(0), stdout);
		tot += sizeof(n_long);

		/* First send the Y */
		if (dump_) 
			fwrite(Y_frame_data_codes, codeWidth * codeHeight, 1, 
			       stdout);
    		net_->send_video_data(T_VIDEOFRAME, 
				      codeWidth * codeHeight >> 2, 
				      0, Y_frame_data_codes, 0);
		tot += codeWidth * codeHeight;

		/* Now interleave the I and Q codes */
		tmp_iq_ptr = IQ_temp_codes;
		tmp_i_ptr = I_frame_data_codes;
		tmp_q_ptr = Q_frame_data_codes;
		for (i = codeWidth * codeHeight / 8; i; i--) {
			tmp_iq_ptr[0] = tmp_i_ptr[0];
			tmp_iq_ptr[1] = tmp_i_ptr[1];
			tmp_iq_ptr[2] = tmp_q_ptr[0];
			tmp_iq_ptr[3] = tmp_q_ptr[1];
			tmp_iq_ptr += 4;
			tmp_i_ptr += 2;
			tmp_q_ptr += 2;
		}
		if (dump_) {
			fwrite(IQ_temp_codes, codeWidth * codeHeight / 2, 1, 
			       stdout);
			fflush(stdout);
		}
		net_->send_video_data(T_VIDEOFRAME, 
				      codeWidth * codeHeight >> 3, 
				      0x200, IQ_temp_codes, 1);
		tot += codeWidth * codeHeight / 2;
#ifdef STATISTICS
		fprintf(stderr, "min: %d max: %d average: %d\n",
			low_error, max_error, total_error / total_vectors);
#endif
	}
	adapt_frames_stored = 0;
	fflush(stdout);
	return (tot);
}
/*
 *--------------------------------------------------------------
 *
 * convert_CrCb_IQ
 *
 *      Converts the scaled input chrominance buffers into YIQ
 *		space using simple table lookup.
 *
 * Results:
 *      The Cr / Cb arrays are converted to I Q in place
 *--------------------------------------------------------------
 */
void 
IpadVQEncoder::convert_CrCb_IQ(u_char* Cr_I, u_char* Cb_Q, int height, 
			       int width, int bytes_per_line)
{
	int     x, y;
	u_char *Cr_I_ptr;
	u_char *Cb_Q_ptr;
	u_short iq;
	Cr_I_ptr = Cr_I;
	Cb_Q_ptr = Cb_Q;
	for (y = height; y; y--) {
		for (x = width; x; x--) {
			iq = CrCb_to_IQ[*Cr_I_ptr | (*Cb_Q_ptr << 8)];
			*Cr_I_ptr++ = iq & 0xff;
			*Cb_Q_ptr++ = iq >> 8;
		}
		Cr_I_ptr += bytes_per_line - width;
		Cb_Q_ptr += bytes_per_line - width;
	}
}
void 
IpadVQEncoder::adapt_codebook()
{
	int     i, j, iter;
	int     y_err, i_err, q_err;
	int     last_y_err, last_i_err, last_q_err;
	last_y_err = last_i_err = last_q_err = 999999;
	y_err = i_err = q_err = 899999;
	for (iter = 0; iter < 10; iter++) {
		/* Quantize to current codebook */
		if ((iter < 3) || (last_y_err > y_err)) {
			last_y_err = y_err;
			y_err = vector_quantize_block(L_Y_scaled_buf, Y_frame_data_codes,
						      y_codes, codeWidth,
				codeHeight * adapt_frames_stored, outWidth);
		}
		if ((iter < 3) || (last_i_err > i_err)) {
			last_i_err = i_err;
			i_err = vector_quantize_block(Cr_I_scaled_buf, I_frame_data_codes,
						      i_codes, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
		}
		if ((iter < 3) || (last_q_err > q_err)) {
			last_q_err = q_err;
			q_err = vector_quantize_block(Cb_Q_scaled_buf, Q_frame_data_codes,
						      q_codes, codeWidth / 2,
				       codeHeight / 2 * adapt_frames_stored,
						      outWidth / 2);
		}
		if (!quietFlag) {
			fprintf(stderr, "Before %d adaption step errors: Y=%d I=%d Q=%d\n",
				iter + 1, y_err, i_err, q_err);
		}
		if ((iter >= 3) &&
		    (last_y_err <= y_err) &&
		    (last_i_err <= i_err) &&
		    (last_q_err <= q_err)) {
			break;
		}
		/* Calculate errors on a per-block basis */
		if ((iter < 3) || (last_y_err > y_err))
			compute_block_errors(L_Y_scaled_buf, 
					     Y_frame_data_codes,
					     Y_frame_coding_errors, y_codes,
					     codeWidth, 
					     codeHeight * adapt_frames_stored,
					     outWidth);
		if ((iter < 3) || (last_i_err > i_err))
			compute_block_errors(Cr_I_scaled_buf, I_frame_data_codes,
					     I_frame_coding_errors, i_codes,
			codeWidth / 2, codeHeight / 2 * adapt_frames_stored,
					     outWidth / 2);
		if ((iter < 3) || (last_q_err > q_err))
			compute_block_errors(Cb_Q_scaled_buf, Q_frame_data_codes,
					     Q_frame_coding_errors, q_codes,
			codeWidth / 2, codeHeight / 2 * adapt_frames_stored,
					     outWidth / 2);
		/* Compute sums and counts */
		if ((iter < 3) || (last_y_err > y_err))
			adapt_compute_code_sums(L_Y_scaled_buf, Y_frame_data_codes,
						y_code_sums,
						y_code_counts, codeWidth,
				codeHeight * adapt_frames_stored, outWidth);
		if ((iter < 3) || (last_i_err > i_err))
			adapt_compute_code_sums(Cr_I_scaled_buf, I_frame_data_codes,
						i_code_sums,
						i_code_counts, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
		if ((iter < 3) || (last_q_err > q_err))
			adapt_compute_code_sums(Cb_Q_scaled_buf, Q_frame_data_codes,
						q_code_sums,
						q_code_counts, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
		if ((iter < 3) || (last_y_err > y_err))
			recompute_codebook(y_codes, y_code_sums, y_code_counts);
		if ((iter < 3) || (last_i_err > i_err))
			recompute_codebook(i_codes, i_code_sums, i_code_counts);
		if ((iter < 3) || (last_q_err > q_err))
			recompute_codebook(q_codes, q_code_sums, q_code_counts);
		/* Delete under-used codebook entries */
		if ((iter < 3) || (last_y_err > y_err))
			prune_codebook(y_codes, y_code_sums, y_code_counts,
				       y_err, 16, 'Y');
		if ((iter < 3) || (last_i_err > i_err))
			prune_codebook(i_codes, i_code_sums, i_code_counts,
				       i_err / 2, 16, 'I');
		if ((iter < 3) || (last_q_err > q_err))
			prune_codebook(q_codes, q_code_sums, q_code_counts,
				       q_err / 2, 16, 'Q');
		/* Supplement codebook w/new entries */
		if ((iter < 3) || (last_y_err > y_err))
			supplement_codebook(y_codes, L_Y_scaled_buf, Y_frame_coding_errors,
				      y_code_sums, y_code_counts, codeWidth,
					    codeHeight * adapt_frames_stored,
					    outWidth, 'Y');
		if ((iter < 3) || (last_i_err > i_err))
			supplement_codebook(i_codes, Cr_I_scaled_buf, I_frame_coding_errors,
				  i_code_sums, i_code_counts, codeWidth / 2,
				       codeHeight * adapt_frames_stored / 2,
					    outWidth / 2, 'I');
		if ((iter < 3) || (last_q_err > q_err))
			supplement_codebook(q_codes, Cb_Q_scaled_buf, Q_frame_coding_errors,
				  q_code_sums, q_code_counts, codeWidth / 2,
				       codeHeight * adapt_frames_stored / 2,
					    outWidth / 2, 'Q');
	}
	/* One last time to center things */
	y_err = vector_quantize_block(L_Y_scaled_buf, Y_frame_data_codes,
				      y_codes, codeWidth,
				codeHeight * adapt_frames_stored, outWidth);
	i_err = vector_quantize_block(Cr_I_scaled_buf, I_frame_data_codes,
				      i_codes, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
	q_err = vector_quantize_block(Cb_Q_scaled_buf, Q_frame_data_codes,
				      q_codes, codeWidth / 2,
				      codeHeight / 2 * adapt_frames_stored,
				      outWidth / 2);
	if (!quietFlag) {
		fprintf(stderr, "After adaption errors: Y=%d I=%d Q=%d\n",
			y_err, i_err, q_err);
	}
	adapt_compute_code_sums(L_Y_scaled_buf, Y_frame_data_codes,
				y_code_sums,
				y_code_counts, codeWidth,
				codeHeight * adapt_frames_stored, outWidth);
	adapt_compute_code_sums(Cr_I_scaled_buf, I_frame_data_codes,
				i_code_sums,
				i_code_counts, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
	adapt_compute_code_sums(Cb_Q_scaled_buf, Q_frame_data_codes,
				q_code_sums,
				q_code_counts, codeWidth / 2,
			codeHeight / 2 * adapt_frames_stored, outWidth / 2);
	recompute_codebook(y_codes, y_code_sums, y_code_counts);
	recompute_codebook(i_codes, i_code_sums, i_code_counts);
	recompute_codebook(q_codes, q_code_sums, q_code_counts);
}
void 
IpadVQEncoder::adapt_compute_code_sums(u_char* image_buf, u_char* frame_codes,
				       int *sums_buffer, int *counts_buffer,
				       int VQ_width, int VQ_height,
				       int pix_per_line)
{
	u_char *image_ptr;
	u_char *code_ptr;
	int    *sums_ptr;
	int     i, j;
	/* Clear counts and sums */
	memset(sums_buffer, 0, sizeof(y_code_sums));
	memset(counts_buffer, 0, sizeof(y_code_counts));
	code_ptr = frame_codes;
	image_ptr = image_buf;
	for (j = VQ_height; j; j--) {
		for (i = VQ_width; i; i--) {
			sums_ptr = sums_buffer + *code_ptr * VECTOR_BYTES;
			counts_buffer[*code_ptr++]++;
			sums_ptr[0] += image_ptr[0];
			sums_ptr[1] += image_ptr[1];
			sums_ptr[2] += image_ptr[2];
			sums_ptr[3] += image_ptr[3];
			image_ptr += pix_per_line;
			sums_ptr[4] += image_ptr[0];
			sums_ptr[5] += image_ptr[1];
			sums_ptr[6] += image_ptr[2];
			sums_ptr[7] += image_ptr[3];
			image_ptr += pix_per_line;
			sums_ptr[8] += image_ptr[0];
			sums_ptr[9] += image_ptr[1];
			sums_ptr[10] += image_ptr[2];
			sums_ptr[11] += image_ptr[3];
			image_ptr += pix_per_line;
			sums_ptr[12] += image_ptr[0];
			sums_ptr[13] += image_ptr[1];
			sums_ptr[14] += image_ptr[2];
			sums_ptr[15] += image_ptr[3];
			image_ptr -= 3 * pix_per_line - 4;
		}
		image_ptr += pix_per_line * 3;
	}
}
void 
IpadVQEncoder::recompute_codebook(u_char * codebook, int *sums, int *counts)
{
	int     i, j;
	for (i = 0; i < 256; i++) {
		if (counts[i]) {
			for (j = 0; j < 16; j++) {
				codebook[i * 16 + j] = sums[i * 16 + j] / counts[i];
			}
		} else {
			memset(codebook + i * 16, 0, 16);
		}
	}
}
/*
 * Routines for finding the vector closest to a desired vector.
 */
#define CHECK_VEC(i,j) \
        error += squares[cb[i] - v[j]]; \
        if (error >= min_error) continue;
u_char 
IpadVQEncoder::BestCode(u_char* codebook, u_char* vector, int bytes_per_line, 
			u_char good_guess, int* total_error_ptr)
{
	unsigned int i;
	register unsigned long error;
	register unsigned long min_error;
	register unsigned char *cb, *v;
	int     best;
	if (error_tolerance && prev_codes_valid) {
		best = good_guess;
		cb = codebook + VECTOR_BYTES * good_guess;
		v = vector;
		min_error = squares[cb[0] - v[0]] + squares[cb[1] - v[1]] +
		    squares[cb[2] - v[2]] + squares[cb[3] - v[3]];
		v += bytes_per_line;
		cb += 4;
		min_error += squares[cb[0] - v[0]] + squares[cb[1] - v[1]] +
		    squares[cb[2] - v[2]] + squares[cb[3] - v[3]];
		v += bytes_per_line;
		cb += 4;
		min_error += squares[cb[0] - v[0]] + squares[cb[1] - v[1]] +
		    squares[cb[2] - v[2]] + squares[cb[3] - v[3]];
		v += bytes_per_line;
		cb += 4;
		min_error += squares[cb[0] - v[0]] + squares[cb[1] - v[1]] +
		    squares[cb[2] - v[2]] + squares[cb[3] - v[3]];
		if (min_error < error_tolerance) {
			*total_error_ptr += min_error;
			return good_guess;
		}
	} else {
		min_error = 0xfffffff;
	}
	cb = codebook;
	for (i = 0; i < CODE_VECTORS; i++, cb += VECTOR_BYTES) {
		error = 0;
		v = vector;
		CHECK_VEC(0, 0);
		CHECK_VEC(1, 1);
		CHECK_VEC(2, 2);
		CHECK_VEC(3, 3);
		v += bytes_per_line;
		CHECK_VEC(4, 0);
		CHECK_VEC(5, 1);
		CHECK_VEC(6, 2);
		CHECK_VEC(7, 3);
		v += bytes_per_line;
		CHECK_VEC(8, 0);
		CHECK_VEC(9, 1);
		CHECK_VEC(10, 2);
		CHECK_VEC(11, 3);
		v += bytes_per_line;
		CHECK_VEC(12, 0);
		CHECK_VEC(13, 1);
		CHECK_VEC(14, 2);
		CHECK_VEC(15, 3);
		min_error = error;
		best = i;
	}
	*total_error_ptr += min_error;
#ifdef STATISTICS
	if (min_error < low_error)
		low_error = min_error;
	if (min_error > max_error)
		max_error = min_error;
	total_error += error;
	total_vectors++;
#endif
/*fprintf(stderr,"%5d %3d %3d\n",min_error,best,good_guess);*/
	return (best);
}
u_char 
IpadVQEncoder::BestCodeHalfRes(u_char* codebook, u_char* vector, 
			       int bytes_per_line, u_char good_guess)
{
	unsigned int i;
	register unsigned long error;
	register unsigned long min_error;
	register unsigned char *cb, *vectl2, *vectl3, *vectl4;
	int     v0, v1, v2, v3;
	int     best;
	v0 = vector[0];
	v1 = vector[1];
	v2 = vector[bytes_per_line];
	v3 = vector[bytes_per_line + 1];
	if (error_tolerance && prev_codes_valid) {
		best = good_guess;
		cb = codebook + good_guess * 4;
		min_error = squares[v0 - cb[0]] + squares[v1 - cb[1]] +
		    squares[v2 - cb[2]] + squares[v3 - cb[3]];
		if (min_error < error_tolerance) {
			return good_guess;
		}
	} else {
		min_error = 0xfffffff;
	}
	cb = codebook;
	for (i = 0; i < CODE_VECTORS; i++, cb += 4) {
		if (((error = squares[v0 - cb[0]]) < min_error) &&
		    ((error += squares[v1 - cb[1]]) < min_error) &&
		    ((error += squares[v2 - cb[2]]) < min_error) &&
		    ((error += squares[v3 - cb[3]]) < min_error)) {
			min_error = error;
			best = i;
		}
	}
#ifdef STATISTICS
	if (min_error < low_error)
		low_error = min_error;
	if (min_error > max_error)
		max_error = min_error;
	total_error += error;
	total_vectors++;
#endif
	return (best);
}
u_char 
IpadVQEncoder::BestCodeUniform(u_char* codebook, u_char* vector, 
			       int bytes_per_line)
{
	unsigned int i;
	register unsigned char *v;
#ifdef STATISTICS
	register unsigned long error;
#endif
	v = vector;
	i = (int) v[0] + (int) v[1] + (int) v[2] + (int) v[3];
	v += bytes_per_line;
	i += (int) v[0] + (int) v[1] + (int) v[2] + (int) v[3];
	v += bytes_per_line;
	i += (int) v[0] + (int) v[1] + (int) v[2] + (int) v[3];
	v += bytes_per_line;
	i += (int) v[0] + (int) v[1] + (int) v[2] + (int) v[3];
	i >>= 4;
#ifdef STATISTICS
	error = squares[vector[0] - i] + squares[vector[1] - i] +
	    squares[vector[bytes_per_line] - i] + squares[vector[bytes_per_line + 1] - i];
	v = vector;
	error = squares[v[0] - i] + squares[v[1] - i] + squares[v[2] - i] + squares[v[3] - i];
	v += bytes_per_line;
	error += squares[v[0] - i] + squares[v[1] - i] + squares[v[2] - i] + squares[v[3] - i];
	v += bytes_per_line;
	error += squares[v[0] - i] + squares[v[1] - i] + squares[v[2] - i] + squares[v[3] - i];
	v += bytes_per_line;
	error += squares[v[0] - i] + squares[v[1] - i] + squares[v[2] - i] + squares[v[3] - i];
	if (error < low_error)
		low_error = error;
	if (error > max_error)
		max_error = error;
	total_error += error;
	total_vectors++;
#endif
	return i;
}
u_char 
IpadVQEncoder::BestCodeHalfResUniform(u_char* codebook, u_char* vector, 
				      int bytes_per_line)
{
	unsigned int i;
#ifdef STATISTICS
	register unsigned long error;
#endif
	i = ((int) vector[0] + (int) vector[1] +
	     (int) vector[bytes_per_line] + (int) vector[bytes_per_line + 1]) >> 2;
#ifdef STATISTICS
	error = squares[vector[0] - i] + squares[vector[1] - i] +
	    squares[vector[bytes_per_line] - i] +
	    squares[vector[bytes_per_line + 1] - i];
	if (error < low_error)
		low_error = error;
	if (error > max_error)
		max_error = error;
	total_error += error;
	total_vectors++;
#endif
	return i;
}
 /* Return value is only valid for full resolution slow search */
 /* In that case it is the error.  Otherwise -1 is returned */
int
IpadVQEncoder::vector_quantize_block(u_char* src, u_char* dst, 
				     u_char* codebook, int VQ_width, 
				     int VQ_height, int bytes_per_line)
{
	int     i, j;
	int     total_error, row_error;
	total_error = -1;
	if (!halfres_coding) {
		if (uniform_coding) {
			for (j = VQ_height; j; j--) {
				for (i = VQ_width; i; i--) {
					*dst++ = BestCodeUniform(codebook, src, bytes_per_line);
					src += 4;
				}
				src += bytes_per_line * 3;
			}
		} else if (fast_coding) {
			for (j = VQ_height; j; j--) {
				for (i = VQ_width; i; i--) {
					*dst++ = BestCodeFast(codebook, src, bytes_per_line);
					src += 4;
				}
				src += bytes_per_line * 3;
			}
		} else {
			for (j = VQ_height; j; j--) {
				row_error = 0;
				for (i = VQ_width; i; i--) {
					*dst = BestCode(codebook, src, bytes_per_line, *dst,
							&row_error);
					dst++;
					src += 4;
				}
				src += bytes_per_line * 3;
				total_error += row_error / VQ_width / 16;
			}
			total_error /= VQ_height;
		}
	} else {
		if (uniform_coding) {
			for (j = VQ_height; j; j--) {
				for (i = VQ_width; i; i--) {
					*dst++ = BestCodeHalfResUniform(codebook, src,
							    bytes_per_line);
					src += 2;
				}
				src += bytes_per_line;
			}
		} else if (fast_coding) {
			for (j = VQ_height; j; j--) {
				for (i = VQ_width; i; i--) {
					*dst++ = BestCodeHalfResFast(codebook, src, bytes_per_line);
					src += 2;
				}
				src += bytes_per_line;
			}
		} else {
			for (j = VQ_height; j; j--) {
				for (i = VQ_width; i; i--) {
					*dst = BestCodeHalfRes(codebook, src, bytes_per_line, *dst);
					dst++;
					src += 2;
				}
				src += bytes_per_line;
			}
		}
	}
	return total_error;
}
void 
IpadVQEncoder::dump_vectors(char* filename, u_char* image_buf,
			    int VQ_width, int VQ_height, int pix_per_line)
{
	int     i, j;
	u_char *image_ptr;
	FILE   *f;
	f = fopen(filename, "w");
	image_ptr = image_buf;
	for (j = VQ_height; j; j--) {
		for (i = VQ_width; i; i--) {
			fwrite(image_ptr, 4, 1, f);
			fwrite(image_ptr + pix_per_line, 4, 1, f);
			fwrite(image_ptr + 2 * pix_per_line, 4, 1, f);
			fwrite(image_ptr + 3 * pix_per_line, 4, 1, f);
			image_ptr += 4;
		}
		image_ptr += pix_per_line * 3;
	}
	fclose(f);
}
 /* compute_block_errors computes the block by block error values and */
 /* stores them away. */
void 
IpadVQEncoder::compute_block_errors(u_char* image_buf, u_char* image_codes, 
				    int* error_buf, u_char* codebook,
				    int VQ_width, int VQ_height, 
				    int pix_per_line)
{
	int     i, j;
	int     error;
	u_char *image_ptr;
	u_char *image_code_ptr;
	u_char *code_ptr;
	int    *error_buf_ptr;
	u_char *cb;
	image_ptr = image_buf;
	image_code_ptr = image_codes;
	error_buf_ptr = error_buf;
	for (j = VQ_height; j; j--) {
		for (i = VQ_width; i; i--) {
			cb = *image_code_ptr++ * 16 + codebook;
			error = squares[image_ptr[0] - cb[0]] +
			    squares[image_ptr[1] - cb[1]] +
			    squares[image_ptr[2] - cb[2]] +
			    squares[image_ptr[3] - cb[3]];
			image_ptr += pix_per_line;
			error += squares[image_ptr[0] - cb[4]] +
			    squares[image_ptr[1] - cb[5]] +
			    squares[image_ptr[2] - cb[6]] +
			    squares[image_ptr[3] - cb[7]];
			image_ptr += pix_per_line;
			error += squares[image_ptr[0] - cb[8]] +
			    squares[image_ptr[1] - cb[9]] +
			    squares[image_ptr[2] - cb[10]] +
			    squares[image_ptr[3] - cb[11]];
			image_ptr += pix_per_line;
			error += squares[image_ptr[0] - cb[12]] +
			    squares[image_ptr[1] - cb[13]] +
			    squares[image_ptr[2] - cb[14]] +
			    squares[image_ptr[3] - cb[15]];
			image_ptr -= 3 * pix_per_line - 4;
			*error_buf_ptr++ = error;
		}
		image_ptr += pix_per_line * 3;
	}
}
void 
IpadVQEncoder::prune_codebook(u_char* codebook, int* sums, int* counts,
			      int ave_limit, int max_limit, char cb_name)
{
	int     i, j, k, s;
	int     ave_error;
	int     max_error;
	int     min_ave;
	int     min_max;
	int     min_ave_index;
	int     min_max_index;
	u_char *cb_i, *cb_j;
	int     max_selfsim[CODE_VECTORS];
	int     ave_selfsim[CODE_VECTORS];
	int     max_selfsim_index[CODE_VECTORS];
	int     ave_selfsim_index[CODE_VECTORS];
	int     unused_entries;
	unused_entries = 0;
	/* First determine codebook codebook statistics */
	for (i = 0; i < CODE_VECTORS; i++) {
		if (counts[i] == 0) {
			unused_entries++;
			continue;
		}
		cb_i = codebook + i * 16;
		max_selfsim[i] = 0xfffffff;
		ave_selfsim[i] = 0xfffffff;
		for (j = 0; j < CODE_VECTORS; j++) {
			if ((i == j) || (counts[j] == 0)) {
				continue;
			}
			cb_j = codebook + j * 16;
			ave_error = max_error = 0;
			for (k = 0; k < 16; k++) {
				s = squares[cb_i[k] - cb_j[k]];
				ave_error += s;
				max_error = max(max_error, s);
			}
			ave_error >>= 4;
			if (ave_error < ave_selfsim[i]) {
				ave_selfsim[i] = ave_error;
				ave_selfsim_index[i] = j;
			}
			if (max_error < max_selfsim[i]) {
				max_selfsim[i] = max_error;
				max_selfsim_index[i] = j;
			}
		}
/*		if (cb_name == 'Y')
		{
			fprintf(stderr,
 			  "%c %3d: Ave with %3d = %5d    Max with %3d = %5d\n",
			  cb_name, i, ave_selfsim_index[i], ave_selfsim[i],
			  max_selfsim_index[i], max_selfsim[i]);
		}
*/
		if ((max_selfsim[i] < max_limit) ||
		    (ave_selfsim[i] < ave_limit)) {
/*if (cb_name=='Y')
			fprintf(stderr,
		"REMOVE %c %3d: Ave with %3d = %5d    Max with %3d = %5d\n",
			cb_name, i, ave_selfsim_index[i], ave_selfsim[i],
					max_selfsim_index[i], max_selfsim[i]);
*/
			counts[i] = 0;
			unused_entries++;
		}
	}
	if (!quietFlag) {
		fprintf(stderr, "%3d %c entries unused\n", unused_entries, 
			cb_name);
	}
}
void 
IpadVQEncoder::supplement_codebook(u_char* codebook, u_char* image_buf,
				   int* coding_errors, int* sums, int* counts,
				   int VQ_width, int VQ_height, 
				   int pix_per_line, char cb_name)
{
	int     i, j, k;
	int     cb_index;
	int     worst_error;
	int     worst_index;
	int     VQ_size = VQ_width * VQ_height;
	int    *error_ptr;
	u_char *cb;
	u_char *image_ptr;
	int     new_error;
	/* Look for a free codebook entry */
	for (cb_index = 0; cb_index < CODE_VECTORS; cb_index++) {
		if (counts[cb_index] != 0) {
			continue;
		}
		/* Now look for most mis-coded block */
		worst_error = coding_errors[0];
		worst_index = 0;
		error_ptr = coding_errors + 1;
		for (j = 1; j < VQ_size; j++, error_ptr++) {
			if (worst_error < *error_ptr) {
				worst_error = *error_ptr;
				worst_index = j;
			}
		}
		/* Fill in codebook entry with it */
		cb = codebook + cb_index * 16;
		image_ptr = worst_index / VQ_width * pix_per_line * 4 +
		    worst_index % VQ_width * 4 +
		    image_buf;
		cb[0] = image_ptr[0];
		cb[1] = image_ptr[1];
		cb[2] = image_ptr[2];
		cb[3] = image_ptr[3];
		cb[4] = image_ptr[0 + pix_per_line];
		cb[5] = image_ptr[1 + pix_per_line];
		cb[6] = image_ptr[2 + pix_per_line];
		cb[7] = image_ptr[3 + pix_per_line];
		cb[8] = image_ptr[0 + 2 * pix_per_line];
		cb[9] = image_ptr[1 + 2 * pix_per_line];
		cb[10] = image_ptr[2 + 2 * pix_per_line];
		cb[11] = image_ptr[3 + 2 * pix_per_line];
		cb[12] = image_ptr[0 + 3 * pix_per_line];
		cb[13] = image_ptr[1 + 3 * pix_per_line];
		cb[14] = image_ptr[2 + 3 * pix_per_line];
		cb[15] = image_ptr[3 + 3 * pix_per_line];
		/* Recompute errors by trying new code on all blocks */
		/* NOTE: we do not update codebook index, sums, or count */
		image_ptr = image_buf;
		error_ptr = coding_errors;
		for (j = VQ_height; j; j--) {
			for (i = VQ_width; i; i--) {
				new_error = squares[image_ptr[0] - cb[0]] +
				    squares[image_ptr[1] - cb[1]] +
				    squares[image_ptr[2] - cb[2]] +
				    squares[image_ptr[3] - cb[3]];
				image_ptr += pix_per_line;
				new_error += squares[image_ptr[0] - cb[4]] +
				    squares[image_ptr[1] - cb[5]] +
				    squares[image_ptr[2] - cb[6]] +
				    squares[image_ptr[3] - cb[7]];
				image_ptr += pix_per_line;
				new_error += squares[image_ptr[0] - cb[8]] +
				    squares[image_ptr[1] - cb[9]] +
				    squares[image_ptr[2] - cb[10]] +
				    squares[image_ptr[3] - cb[11]];
				image_ptr += pix_per_line;
				new_error += squares[image_ptr[0] - cb[12]] +
				    squares[image_ptr[1] - cb[13]] +
				    squares[image_ptr[2] - cb[14]] +
				    squares[image_ptr[3] - cb[15]];
				image_ptr -= 3 * pix_per_line - 4;
				if (new_error < *error_ptr) {
					*error_ptr = new_error;
					counts[cb_index]++;
				}
				error_ptr++;
			}
			image_ptr += pix_per_line * 3;
		}
/*if (cb_name=='Y')
		fprintf(stderr,"Entry %c %3d now has %d uses\n",
			cb_name, cb_index, counts[cb_index]);
*/
	}
}
