/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
#ifndef lint
static char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/vgw/ipad/RCS/vq_code.cc,v 1.4 1995/11/22 03:28:25 elan Exp $ (UC Berkeley)";
#endif

#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>

#include "encoder-ipadvq.h"

#define SOLID_LEVELS    56
#define VARIATIONS ((256 - SOLID_LEVELS) / 8)

#ifndef min
#define min(a,b)  (((a)<(b)) ? (a) : (b))
#endif

/* Sum of num_levels should be VARIATIONS */
int num_levels[3][5] = {{1, 3, 5, 7, 9},
		        {1, 3, 5, 7, 9},
		        {1, 3, 5, 7, 9}};
int min_levels[3][5] = {{0, 0, 0, 0, 0},
			{64, 64, 64, 64, 64},
			{64, 64, 64, 64, 64}};
int max_levels[3][5] = {{255, 255, 255, 255, 255},
			{200, 200, 200, 200, 200},
			{200, 200, 200, 200, 200}};    
void
IpadVQEncoder::generate_fast_codebooks(u_char* y_codes, u_char* i_codes, 
				       u_char* q_codes)
{
	unsigned char codebook[3][256][4][4];
	int     i, j, k;
	int     cbook;
	int     hi, lo;
	int     levels;
	int     x, y;
	int     index;
	int     error;
	int     min_error;
	int     best;
	int     v0, v1, v2, v3;
	int     codeV0[256], codeV1[256], codeV2[256], codeV3[256];
	long   *sq = &cb_squares[256];
	for (i = -256; i < 256; i++) {
		cb_squares[i + 256] = i * i;
	}
	y_codes_addr = y_codes;
	i_codes_addr = i_codes;
	q_codes_addr = q_codes;
	memset(codebook, 0, sizeof(codebook));
	for (cbook = 0; cbook < 3; cbook++) {
		for (index = 0; index < SOLID_LEVELS; index++) {
			for (y = 0; y < 4; y++) {
				for (x = 0; x < 4; x++) {
					codebook[cbook][index][y][x] =
					    index * 256 / SOLID_LEVELS;
				}
			}
		}
		for (i = 0; i < sizeof(num_levels[0]) / sizeof(int); i++) {
			for (j = 0; j < num_levels[cbook][i]; j++) {
				lo = min_levels[cbook][i] +
				    (max_levels[cbook][i]
				     - min_levels[cbook][i])
				    * j / num_levels[cbook][i];
				hi = min_levels[cbook][i] +
				    (max_levels[cbook][i]
				     - min_levels[cbook][i])
				    * (j + 1) / num_levels[cbook][i];
				for (y = 0; y < 4; y++) {
					for (x = 0; x < 4; x++) {
#define SMOOTH
#ifdef SMOOTH
						codebook[cbook][index][y][x] =
							((hi * x) + lo * (3 - x)) / 3;
						codebook[cbook][index + VARIATIONS]
						    [y][x] =
						    ((lo * x) + hi * (3 - x)) / 3;
						codebook[cbook][index + 2 * VARIATIONS]
						    [y][x] =
						    ((hi * y) + lo * (3 - y)) / 3;
						codebook[cbook][index + 3 * VARIATIONS]
						    [y][x] =
						    ((lo * y) + hi * (3 - y)) / 3;
						codebook[cbook][index + 4 * VARIATIONS]
						    [y][x] =
						    ((hi * (x + y)) + lo * (6 - x - y)) / 6;
						codebook[cbook][index + 5 * VARIATIONS]
						    [y][x] =
						    ((lo * (x + y)) + hi * (6 - x - y)) / 6;
						codebook[cbook][index + 6 * VARIATIONS]
						    [y][x] =
						    ((hi * (x - y + 3)) + lo * (3 - x + y)) / 6;
						codebook[cbook][index + 7 * VARIATIONS]
						    [y][x] =
						    ((lo * (x - y + 3)) + hi * (3 - x + y)) / 6;
#else
						codebook[cbook][index][y][x] =
						    x >= 2 ? hi : lo;
						codebook[cbook][index + VARIATIONS]
						    [y][x] =
						    x >= 2 ? lo : hi;
						codebook[cbook][index + 2 * VARIATIONS]
						    [y][x] =
						    y >= 2 ? hi : lo;
						codebook[cbook][index + 3 * VARIATIONS]
						    [y][x] =
						    y >= 2 ? lo : hi;
						codebook[cbook][index + 4 * VARIATIONS]
						    [y][x] =
						    x + y > 3 ? hi : lo;
						codebook[cbook][index + 5 * VARIATIONS]
						    [y][x] =
						    x + y > 3 ? lo : hi;
						codebook[cbook][index + 6 * VARIATIONS]
						    [y][x] =
						    x - y > 0 ? hi : lo;
						codebook[cbook][index + 7 * VARIATIONS]
						    [y][x] =
						    x - y > 0 ? lo : hi;
#endif
					}
				}
				index++;
			}
		}
	}
	memcpy(y_codes, codebook[0], 4096);
	memcpy(i_codes, codebook[1], 4096);
	memcpy(q_codes, codebook[2], 4096);
	/* Now make quick-map for combos */
	for (cbook = 0; cbook < 3; cbook++) {
		for (index = 0; index < 256; index++) {
			codeV0[index] = ((int) codebook[cbook][index][0][0] +
					 (int) codebook[cbook][index][0][1] +
					 (int) codebook[cbook][index][1][0] +
				   (int) codebook[cbook][index][1][1]) >> 2;
			codeV1[index] = ((int) codebook[cbook][index][0][2] +
					 (int) codebook[cbook][index][0][3] +
					 (int) codebook[cbook][index][1][2] +
				   (int) codebook[cbook][index][1][3]) >> 2;
			codeV2[index] = ((int) codebook[cbook][index][2][0] +
					 (int) codebook[cbook][index][2][1] +
					 (int) codebook[cbook][index][3][0] +
				   (int) codebook[cbook][index][3][1]) >> 2;
			codeV3[index] = ((int) codebook[cbook][index][2][2] +
					 (int) codebook[cbook][index][2][3] +
					 (int) codebook[cbook][index][3][2] +
				   (int) codebook[cbook][index][3][3]) >> 2;
		}
		/* First group is solids.  Indexed by full code */
		for (i = 0; i < 256; i++) {
			min_error = 0xfffffff;
			for (index = 0; index < SOLID_LEVELS; index++) {
				error = sq[codeV0[index] - i] + sq[codeV1[index] - i] +
				    sq[codeV2[index] - i] + sq[codeV3[index] - i];
				if (min_error > error) {
					min_error = error;
					best = index;
				}
			}
			solid_2_code[cbook][i] = best;
		}
		/* Second group is horizontal gradient */
		for (i = 0; i < 255; i += 4) {
			for (j = 0; j < 255; j += 4) {
				min_error = 0xfffffff;
				for (index = SOLID_LEVELS;
				     index < SOLID_LEVELS + VARIATIONS * 2;
				     index++) {
					error = sq[codeV0[index] - i] + sq[codeV1[index] - j] +
					    sq[codeV2[index] - i] + sq[codeV3[index] - j];
					if (min_error > error) {
						min_error = error;
						best = index;
					}
				}
				horiz_2_code[cbook][i >> 2][j >> 2] = best;
			}
		}
		/* Third group is vertical gradient */
		for (i = 0; i < 255; i += 4) {
			for (j = 0; j < 255; j += 4) {
				min_error = 0xfffffff;
				for (index = SOLID_LEVELS + VARIATIONS * 2;
				     index < SOLID_LEVELS + VARIATIONS * 4;
				     index++) {
					error = sq[codeV0[index] - i] + sq[codeV1[index] - i] +
					    sq[codeV2[index] - j] + sq[codeV3[index] - j];
					if (min_error > error) {
						min_error = error;
						best = index;
					}
				}
				vert_2_code[cbook][i >> 2][j >> 2] = best;
			}
		}
		/* Forth group is vertical gradient */
		for (i = 0; i < 255; i += 4) {
			for (j = 0; j < 255; j += 4) {
				k = (i + j) >> 1;
				min_error = 0xfffffff;
				for (index = SOLID_LEVELS + VARIATIONS * 4;
				     index < SOLID_LEVELS + VARIATIONS * 6;
				     index++) {
					error = sq[codeV0[index] - i] + sq[codeV1[index] - k] +
					    sq[codeV2[index] - k] + sq[codeV3[index] - j];
					if (min_error > error) {
						min_error = error;
						best = index;
					}
				}
				diag0_2_code[cbook][i >> 2][j >> 2] = best;
			}
		}
		/* Fiftth group is vertical gradient */
		for (i = 0; i < 255; i += 4) {
			for (j = 0; j < 255; j += 4) {
				k = (i + j) >> 1;
				min_error = 0xfffffff;
				for (index = SOLID_LEVELS + VARIATIONS * 6;
				     index < SOLID_LEVELS + VARIATIONS * 8;
				     index++) {
					error = sq[codeV0[index] - k] + sq[codeV1[index] - i] +
					    sq[codeV2[index] - j] + sq[codeV3[index] - k];
					if (min_error > error) {
						min_error = error;
						best = index;
					}
				}
				diag1_2_code[cbook][i >> 2][j >> 2] = best;
			}
		}
	}
}
int
IpadVQEncoder::BestCodeFast(u_char* codebook, u_char* vector, 
			    int bytes_per_line)
{
	unsigned int i, j;
	register unsigned long error;
	register unsigned long min_error;
	register long *sq;
	register unsigned char *cb, *vectl2, *vectl3, *vectl4;
	int     v0, v1, v2, v3;
	int     best;
	int     cnum;
	if (codebook == i_codes_addr) {
		cnum = 1;
	} else if (codebook == q_codes_addr) {
		cnum = 2;
	} else {
		cnum = 0;
	}
	sq = &cb_squares[256];
	vectl2 = vector + bytes_per_line;
	vectl3 = vectl2 + bytes_per_line;
	vectl4 = vectl3 + bytes_per_line;
	v0 = ((int) vector[0] + (int) vector[1] +
	      (int) vectl2[0] + (int) vectl2[1]) >> 2;
	v1 = ((int) vector[2] + (int) vector[3] +
	      (int) vectl2[2] + (int) vectl2[3]) >> 2;
	v2 = ((int) vectl3[0] + (int) vectl3[1] +
	      (int) vectl4[0] + (int) vectl4[1]) >> 2;
	v3 = ((int) vectl3[2] + (int) vectl3[3] +
	      (int) vectl4[2] + (int) vectl4[3]) >> 2;
	/* First get error for uniform coding */
	best = solid_2_code[cnum][(v0 + v1 + v2 + v3) >> 2];
	i = codebook[best << 2];
	min_error = sq[v0 - i] + sq[v1 - i] + sq[v2 - i] + sq[v3 - i];
	/* Next for horiz gradient */
	i = horiz_2_code[cnum][(v0 + v2) >> 3][(v1 + v3) >> 3];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for vert gradient */
	i = vert_2_code[cnum][(v0 + v1) >> 3][(v2 + v3) >> 3];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for diag0 gradient */
	i = diag0_2_code[cnum][(v0 * 2 + v1 + v2) >> 4][(v3 * 2 + v1 + v2) >> 4];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for diag1 gradient */
	i = diag1_2_code[cnum][(v1 * 2 + v0 + v3) >> 4][(v2 * 2 + v0 + v3) >> 4];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	return best;
}

int
IpadVQEncoder::BestCodeHalfResFast(u_char* codebook, u_char* vector, 
				   int bytes_per_line)
{
	unsigned int i, j;
	register unsigned long error;
	register unsigned long min_error;
	register long *sq;
	int     v0, v1, v2, v3;
	int     best;
	int     cnum;
	if (codebook == i_codes_addr) {
		cnum = 1;
	} else if (codebook == q_codes_addr) {
		cnum = 2;
	} else {
		cnum = 0;
	}
	sq = &cb_squares[256];
	v0 = vector[0];
	v1 = vector[1];
	v2 = vector[bytes_per_line];
	v3 = vector[bytes_per_line + 1];
	/* First get error for uniform coding */
	best = solid_2_code[cnum][(v0 + v1 + v2 + v3) >> 2];
	i = codebook[best << 2];
	min_error = sq[v0 - i] + sq[v1 - i] + sq[v2 - i] + sq[v3 - i];
	/* Next for horiz gradient */
	i = horiz_2_code[cnum][(v0 + v2) >> 3][(v1 + v3) >> 3];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for vert gradient */
	i = vert_2_code[cnum][(v0 + v1) >> 3][(v2 + v3) >> 3];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for diag0 gradient */
	i = diag0_2_code[cnum][(v0 * 2 + v1 + v2) >> 4][(v3 * 2 + v1 + v2) >> 4];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	/* Now for diag1 gradient */
	i = diag1_2_code[cnum][(v1 * 2 + v0 + v3) >> 4][(v2 * 2 + v0 + v3) >> 4];
	error = sq[v0 - codebook[i << 2]] + sq[v1 - codebook[(i << 2) + 1]] +
	    sq[v2 - codebook[(i << 2) + 2]] + sq[v3 - codebook[(i << 2) + 3]];
	if (error < min_error) {
		min_error = error;
		best = i;
	}
	return best;
}
