/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 * @(#) $Header: /home/barad-dur/b/users/elan/src/vgw/vgw/ipad/RCS/vq_file_fmt.h,v 1.4 1995/11/22 03:28:25 elan Exp $ (UC Berkeley)
 */

/* vq_file_fmt.h */
/* Jeff Gilbert */
/* 4/9/95 */

#ifndef vgw_vq_file_fmt_h
#define vgw_vq_file_fmt_h

#include <sys/types.h>

/*
 * General Format of VQ (not CM chunked) files (this applies both to infopad 
 * files as well as could be used for other VQ types:
 * 
 * 	uint32		magic_number ("IPVQ-1.0" = 0x49505651)
 * 	uint16		major version number 		(1)
 * 	uint16		minor version number		(0)
 * 	uint32		width (in VQ IQ entries - i.e. 128/8=16 for us
 * 	uint32		height (240/8=30 for us)
 * 	uint32		framesize = width*height*6
 * 	uint32		codebook_entries (768)
 * 	uint32		codebook_entry_size (16)
 * 	uint32		frames per second * 65536  (0 means don't know)
 * 	uint32		Length_of_ignored data
 * 	uchar8 * Length_of_ignored  ignored data
 * 
 * 		Then each frame consists of:
 * 	uint32		entries_to_update	(>=0	< codebook_entries)
 * 
 * 		if non-zero this is followed by:
 * 	uint32		first_entry_num		(>=0   <= codebook_entries -
 * 							  entries_to_update)
 * 
 * 	uchar*entries_to_update*codebook_entry_size  	Codebook Data
 * 	uint32		entries_to_update		Until 0
 * 			...
 * 
 * 
 * 	Then after the 0 entries_to_update comes the data frame data which is
 * 		framesize bytes long (you should ignore width and height and
 * 			if you need a fixed size, force framesize to match)
 * 
 * 	NOTE:  It is nice if you detect a entries_to_update=VQ_MAGIC_NUM and
 * 			realize that that means that two vq files were cat'ed 
 * 			together and just read the new header and continue
 * 
 * Infopad specific VQ files
 * 	codebook_entries must be 768
 * 	codebook_entry_size must be 16
 * 	Width & height must be 16 and 30 to be sent over IP network (for emu
 * 		or hardware).  Other sizes can be transcoded and played.
 * 	Note that the codebook is slightly different than that hardware expects
 * 		but the send program will correct on the fly.    
 * 
 * 	Codebook (total of 768 entries):
 * 		Entries   0..255:		Y Codebook
 * 		Entries 256..511:		I Codebook
 * 		Entries 512..767		Q Codebook
 * 
 * 	Codebook entries (16 bytes apiece):
 * 		A B C D
 * 		E F G H			The 16 data bytes are present
 * 		I J K L			in A..P order
 * 		M N O P
 * 
 * 	Y Codebook values are 0..255 unsigned binary - 0=black, 255=white
 * 	I,Q Codebook values are 0..255 offset binary: 128 means 0, 127 is
 * 			-1, 129 is 1 etc.  Note that the acutal IP
 * 			hardware uses sign magnitude but we do not here as
 * 			we will have to remap for hardware as it only wants
 * 			6 bits and also wants ordered by line then entry so
 * 			just simple reformat.
 * 
 * 	Frame Data:
 * 		Y Frame (Width*height*4 bytes)
 * 		The Y data (codebook entries - 1 byte per 4x4 block) is 
 * 		present in normal scan order - top left to top right
 * 		down to bottom right.
 * 
 * 		IQ Frame (Width*height*2 bytes)
 * 		I and Q are both subsampled by 2.  They are present in
 * 		foursomes of 2 apiece i.e. 
 * 				I  I  Q  Q    I  I  Q  Q    I  I  Q  Q
 * 				0  1  0  1    2  3  2  3    4  5  4  5
 * 
 * 		where the 0,1,2,3.. correspond again to normal scan order
 * 
 *
 */

			/* VQ Magic number "IPVQ" = 0x49505651 */
#define VQ_MAGIC_NUM			0x49505651

/* All numbers should be in net order */

struct VQ_file_header {
	u_long		magic_number;
	u_short		major_version;
	u_short		minor_version;
	u_long		width;
	u_long		height;
	u_long		frame_size;
	u_long		codebook_entries;
	u_long		codebook_entry_size;
	u_long		frames_per_sec;
	u_long		extra_data_len;
};


#endif /* vgw_vq_file_fmt_h */
