#
# Copyright (c) 1995 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/RCS/rtpgw.tcl,v 1.40 1997/03/26 03:46:45 elan Exp $ (UC Berkeley)
#
set encoders(audio) { lpc }
set encoders(video) { h261 }
set codec_map(video) { { jpeg h261 jpeg/dct h261/dct } 
			 { nv h261 nv h261/pixel } 
			 { h261 h261 h261/dct h261/dct } }
set codec_map(audio) { {adpcm lpc adpcm lpc} { pcm lpc pcm lpc } }

set rtp_pt(-1) ""
set rtp_pt(0) pcm
set rtp_pt(1) celp
set rtp_pt(3) gsm
set rtp_pt(5) adpcm
set rtp_pt(7) lpc
set rtp_pt(20) ldct
set rtp_pt(21) skid
set rtp_pt(22) bvc
set rtp_pt(23) rgb8
set rtp_pt(25) cellb
set rtp_pt(26) jpeg
set rtp_pt(27) cuseeme
set rtp_pt(28) nv
set rtp_pt(29) picw
set rtp_pt(30) cpv
set rtp_pt(31) h261
set rtp_pt(32) mpeg
set rtp_pt(33) mpeg2

set active_list {}

proc register src {
}

proc unregister src {
}

proc change_name { src } {
# XXX not yet
#	set name [$src name]
#	global nametag srclist
#	if [info exists srclist] {
#		$srclist itemconfigure $nametag($src) -text $name
#	}

	update_source_info $src
}

proc embolden src {
}

proc disable src {
}

#
# Return true if the two strings provide more or less the
# same informational content
#
proc cname_redundant { name cname } {
	set ni [string first @ $name]
	if { $ni < 0 } {
		return 0
	}
	set ci [string first @ $cname]
	if { $ci < 0 } {
		return 0
	}
	if { [string compare \
		[string range $name 0 $ni] \
		[string range $cname 0 $ci]] == 0 } {
		return 1
	}
	return 0
}

proc update_source_info { src } {
	global srcinfo rtp_pt
	#
	# Figure out best presentation from available information.
	#
	set name [$src sdes name]
	set cname [$src sdes cname]
	set addr [$src addr]
	if { $name == "" } {
		if { $cname == "" } {
			set srcinfo($src,name) $addr
			set srcinfo($src,info) $addr/$rtp_pt([$src ifmt])
		} else {
			set srcinfo($src,name) $cname
			set srcinfo($src,info) $addr/$rtp_pt([$src ifmt])
		}
	} elseif [cname_redundant $name $cname] {
		set srcinfo($src,name) $name
		set srcinfo($src,info) $addr/$rtp_pt([$src ifmt])
	} else {
		set srcinfo($src,name) $name
		set srcinfo($src,info) $cname/$rtp_pt([$src ifmt])
	}
}

proc activate { src } {
	global active_list srcinfo V

	lappend active_list $src
	change_format $src

	set addr [$src addr]	
	set t [$src transcoder]
	set tname [$t tname]
	set osess [$src osess]
 	cbBroadcast "activate $src {$srcinfo($src,name)} \
				   {$addr} \
				   $osess \
				   {$srcinfo($src,info)} \
				   {$tname} "
}

proc deactivate src {
    	global active_list 
	set i [lsearch $active_list $src]
	set active_list [lreplace $active_list $i $i]
	if { [info exists ofpshat($src)] } {
		unset ifpshat($src)
		unset ibpshat($src)
		unset ofpshat($src)
		unset obpshat($src)
	}
    	cbBroadcast "deactivate $src"
}

proc change_format { src } {
	global V

	set osess $V([$src osess])
	set rc [$osess rate-control]
       	
	if { $rc == 0 } {
		set t [new transcoder null]
		$t tname "pass-thru"
		$t forward [expr 1-$rc]
	} else {
		global rtp_pt
		set ifmt $rtp_pt([$src ifmt])			
		set t [trans_match $ifmt [$osess ofmt]]
	}
	
	$src transcoder $t

	# XXX Should be transcoder attribute.
	if { [$t media-type] == "video" } {
		$src color [$osess color]
	}

	set_rate_vars $src
	update_source_info $src
}

proc grayout src {
}

proc set_rate_vars { src } {
	global ofpshat obpshat ibpshat ifpshat
	if { [info exists ofpshat($src)] } {
		unset ifpshat($src)
		unset ibpshat($src)
		unset ofpshat($src)
		unset obpshat($src)
	}

	set gain [option get filterGain]

	set ifpshat($src) 0
	rate_variable ifpshat($src) $gain
	set ibpshat($src) 0
	rate_variable ibpshat($src) $gain
	set ofpshat($src) 0
	rate_variable ofpshat($src) $gain
	set obpshat($src) 0
	rate_variable obpshat($src) $gain
}

proc srcupdate { srcspec src } {
	global ifpshat ibpshat ofpshat obpshat active_list srcinfo V

	# May have had a message in transit.
    	if { [lsearch $active_list $src] == -1 } then {
		return
	}

	set ifpshat($src) [$src nf]
	set ibpshat($src) [$src nb]
	set ofpshat($src) [$src ofrms]
	set obpshat($src) [$src obits]

	set ifps $ifpshat($src)
	set ibps $ibpshat($src)
	set ofps $ofpshat($src)
	set obps $obpshat($src)
	
	set ifps [format_fps $ifps]
	set ibps [format_bps $ibps]
	set ofps [format_fps $ofps]
	set obps [format_bps $obps]

	set maxbps [$src bps]
	set t [$src transcoder]
	set tname [$t tname]
	set osess [$src osess]
    	cbSend $srcspec "srcupdate $src {$srcinfo($src,name)} \
					$osess \
				        {$srcinfo($src,info)} \
				        {$tname} \
				        $maxbps \
					{{$ifps} {$ibps} {$ofps} {$obps}} "
}

proc format_fps f {
	set fps $f
	if { $fps < .1 } {
		set fps "0 f/s"
	} elseif { $fps < 10 } {
		set fps [format "%.1f f/s" $fps]
	} else {
		set fps [format "%2.0f f/s" $fps]
	}

	return $fps
}

proc format_bps b {
	set bps $b

	if { $bps < 1 } {
		set bps "0 bps"
	} elseif { $bps < 1000 } {
		set bps [format "%3.0f bps" $bps]
	} elseif { $bps < 1000000 } {
		set bps [format "%3.0f kb/s" [expr $bps / 1000.]]
	} else {
		set bps [format "%.1f Mb/s" [expr $bps / 1000000.]]
	}
	
	return $bps
}

proc init_confbus {channel} {
	if { $channel != 0 } {
		global confbus
		set confbus [new confbus $channel]
		$confbus handler confbusHandler
		global V
		set V(cb_channel) $channel
		init_callbacks
	}
}

proc appInit { app type channel } {
	global V cbAppName encoders codec_map
    
	set V(app) $app
	set V(sm) [new session-manager $type/rtp]
    	set V(source-manager) [new source-manager $type]
	set V(A) [$V(sm) A-session]
	set V(B) [$V(sm) B-session]
	set V(codec_map) $codec_map($type)
	set V(encoders) $encoders($type)

	set cbAppName $app

	init_confbus $channel
	
	alive
}	
	
proc init_callbacks {} {
	cb_callback cb_init { cb_init }
	cb_callback cb_get { cb_get }
	cb_callback cb_set { cb_set }
	cb_callback cb_video_set { cb_video_set }
	cb_callback cb_srcset { cb_srcset }
	cb_callback srcupdate { srcupdate }
	cb_callback src_switch { src_switch }
	cb_callback cb_crypt_set { cb_crypt_set }
}

proc cb_get { srcspec var } {
	global V
	switch $var {
		encoders {
			cbSend $srcspec "encoders {$V(encoders)}"
		}
		active_list {
			global active_list
			cbSend $srcspec "active_list {$active_list}"
		}
	}
}

proc cb_init { srcspec } {
	global active_list V

	foreach s $active_list {
		$s deactivate
	}
	$V(sm) init
}	

proc session_addr { sname dst } {
	global V

	# XXX Should initialize sessions addr structures

	set dstlist [split $dst :]
	foreach dst $dstlist {
		add_dst $sname $dst
	}

	$V($sname) probe-init
}

proc add_dst {sname dst} {
	global V

	set dst [split $dst /]

	set addr [lindex $dst 0]
	set port [lindex $dst 1]
	set ttl  [lindex $dst 2]
    
	
	set datanet [new network ip]
	$datanet open $addr $port $ttl
	$V($sname) data-net $datanet
	set V($sname,data-net) $datanet

    	incr port
	
	set ctrlnet [new network ip]
	$ctrlnet open $addr $port $ttl
	$V($sname) ctrl-net $ctrlnet
	set V($sname,ctrl-net) $ctrlnet

    	incr port
	
	set probenet [new network ip]
	$probenet open $addr $port $ttl
	$V($sname) loop-probe-net $probenet
}

proc cb_set { srcspec sname var val } {
	global V active_list
	switch $var {
		addr {
		    	session_addr $sname $val
		} 
		RC {
			$V($sname) rate-control $val
			foreach s $active_list {
				if { [$s osess] != $sname } {
					continue
				}			    
				if { $val == 1 } then {
					# rate control
					global rtp_pt
					set ifmt $rtp_pt([$s ifmt])
					set ofmt [$V($sname) ofmt]
					set newt [trans_match $ifmt $ofmt]
					$newt forward 0
				} else {
					# no rate control
					set newt [new transcoder null]
					$newt tname "pass-thru"
					$newt forward 1
				}
				$s transcoder $newt
				# XXX
				if { [$newt media-type] == "video" } {
					$s color [$V($sname) color]
				}
				set_rate_vars $s
			}
		} 
		ofmt {
			$V($sname) ofmt $val
			set rc [$V($sname) rate-control]
			# If no rate control, no need to do anything.
			if { $rc == 0 } then {
				return
			}
			foreach s $active_list {
				if { [$s osess] != $sname } {
					continue
				}
				global rtp_pt
				set ifmt $rtp_pt([$s ifmt])
				set newt [trans_match $ifmt $val]
				$newt forward 0
				$s transcoder $newt
				# XXX
				if { [$newt media-type] == "video" } {
					$s color [$V($sname) color]
				}
				set_rate_vars $s
			}
		}
		default {
			puts "$V(app): cb_set: unknown variable \"$var\"
"
		}
	}
}

proc cb_video_set { srcspec sname var val } {
	global V active_list
	switch $var {
		color {
		    	foreach s $active_list {
				if { [$s osess] != $sname } {
					continue
				}
				set t [$s transcoder]
				if {[$t tname] != "pass-thru"} {
					$s color $val
				}
				$V($sname) color $val
			}
		}
		default {
			cb_set $srcspec $sname $var $val
		}
	}
}

proc cb_audio_set { srcspec sname var val } {
	global V  active_list
	switch $var {
		default {
			cb_set $srcspec $sname $var $val
		}
	}
}

proc set_padid { cmd v } {
	puts "$V(app): Engine not configured for Infopad!"	
}

proc cb_srcset { srcspec src type val } {
	global V active_list

	# May have had a message in transit.
    	if { [lsearch $active_list $src] == -1 } then {
		return
	}
	switch $type {
		maxbps {
			$src bps $val
		}
	}
}

proc cb_callback { proc methods } {
	global cb_dispatch
	foreach i $methods {
		set cb_dispatch($i) $proc
	}
}

proc trans_match { ifmt ofmt } {
	global V

    	foreach e $V(codec_map) {
		set m_ifmt [lindex $e 0]
		set m_ofmt [lindex $e 1]
		if { $m_ifmt == $ifmt && $m_ofmt == $ofmt } then {
			set encoder [new encoder [lindex $e 3]]
			set t [new transcoder [lindex $e 2]]
			$t encoder $encoder
			$t tname "$ifmt->$ofmt"
			return $t
		}
	}
	set t [new transcoder null]
	$t tname "pass-thru"
	return $t
}

proc alive {} {
	cbBroadcast "alive"
	after 10000 alive
}

proc adios {} {
	cbBroadcast "dead"
}

proc loopdetect {msg} {
    	cbBroadcast "loopdetect {$msg}"
    	adios
}

# XXX
proc src_switch addr {
}
