/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/RCS/session.cc,v 1.27 1996/06/14 22:05:37 elan Exp $ (LBL)";

#include <osfcn.h>
#include <math.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include "session.h"
#include "net.h"
#include "transmitter.h"
#include "source-gw.h"
#include "ntp-time.h"


#ifdef INFOPAD
#include "ipadchan.h"
#endif

#if defined(sun) && !defined(__svr4__)
extern "C" {
extern long random();
extern int gettimeofday(struct timeval*, struct timezone*);
}
#endif

static class SessionManagerMatcher : public Matcher {
public:
	SessionManagerMatcher() : Matcher("session-manager") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "audio/rtp") == 0)
			return (new AudioSessionManager);
		else if (strcmp(id, "video/rtp") == 0)
			return (new VideoSessionManager);
		else
			return (0);
	}
} session_manager_matcher;

int VideoSessionManager::check_format(int fmt)
{
	switch(fmt) {
	case RTP_PT_CELLB:
	case RTP_PT_JPEG:
	case RTP_PT_CUSEEME:
	case RTP_PT_NV:
	case RTP_PT_CPV:
	case RTP_PT_H261:
	case RTP_PT_BVC:
	case RTP_PT_SKID:
	case RTP_PT_LDCT:
	case RTP_PT_H261_COMPAT:/*XXX*/
		return (1);
	}
	return (0);
}

int AudioSessionManager::check_format(int fmt)
{
	switch (fmt) {
	case RTP_PT_PCMU:
	case RTP_PT_CELP:
	case RTP_PT_GSM:
	case RTP_PT_DVI:
	case RTP_PT_LPC:
		return (1);
	}
	return (0);
}

extern SIGRET adios(...);

void DataHandler::dispatch(int) 
{
	sm_.recv(this);
}

void CtrlHandler::dispatch(int) 
{
	sm_.recv(this);
}

void LoopProbeHandler::dispatch(int) 
{
	sm_.recv(this);
}


SessionManager::SessionManager()
	       : last_np_(0), 
		 rtcp_avg_size_(128.), badversion_(0), badoptions_(0), 
		 nrunt_(0), badfmt_(0), A_("A", this), B_("B", this)
{}

void SessionManager::recv(DataHandler* dh)
{
	n_long src;
	/* leave room in case we need to expand rtpv1 into an rtpv2 header */
	u_char* bp = &pktbuf_[4];
	Network* n = dh->net();
	int cc = n->recv(bp, 2 * RTP_MTU - 4, src);

	if (cc <= 0)
		return;

	/* 
	 * Determine if we can accept the packet. 
	 */
	if (A_.dh() == 0 || B_.dh() == 0)
		return;

	/* 
	 * Process packet.
	 */
	rtphdr* rh = (rtphdr*)bp;
	int version = *(u_char*)rh >> 6;
	if (version != 2) {
		++badversion_;
		return;
	}
	
	bp += sizeof(*rh);
	cc -= sizeof(*rh);

	if (cc < 0) {
		++nrunt_;
		return;
	}
	n_long srcid = rh->rh_ssrc;
	int flags = ntohs(rh->rh_flags);
	if ((flags & RTP_X) != 0) {
		++badext_;
		return;
	}

	/*
	 * Check for illegal payload types.  Most likely this is
	 * a session packet arriving on the data port.
	 */
	int fmt = flags & 0x7f;
	if (!check_format(fmt)) {
		++badfmt_;
		return;
	}

	SourceManager& sm = gwSourceManager::instance();
	u_int16_t seqno = ntohs(rh->rh_seqno);
	gwSource* s = (gwSource*)sm.demux(srcid, src, seqno);
	if (s == 0)
		/*
		 * Takes a pair of validated packets before we will
		 * believe the source.  This prevents a runaway
		 * allocation of Source data structures for a
		 * stream of garbage packets.
		 */
		return;
	
	timeval now = unixtime();
	s->lts_data(now);
	s->sts_data(rh->rh_ts);
	
	/*
	 * This is a data packet.  If the source needs activation,
	 * or the packet format has changed, deal with this.
	 * Then, hand the packet off to the packet handler.
	 * XXX might want to be careful about flip-flopping
	 * here when format changes due to misordered packets
	 * (easy solution -- keep rtp seqno of last fmt change).
	 */
	Network* dst;
	const char* osess;
	if (dh->sess() == &B_) {
		dst = A_.datanets();
		osess = A_.sname();
	} else {
		dst = B_.datanets();
		osess = B_.sname();
	}

	PacketHandler* h = s->handler();
	if (h == 0) {
		h = s->activate(fmt, dst, osess);
	} else if (s->format() != fmt)
		h = s->change_format(fmt);

	/*
	 * XXX bit rate doesn't include rtpv1 options;
	 * but v1 is going away anyway.
	 */
	int dup = s->cs(seqno);
	s->np(1);
	s->nb(cc + sizeof(*rh));
	if (dup)
		return;

	int hlen = h->hdrlen();
	if (cc < hlen) {
		s->runt(1);
		return;
	}

	if (h->recv(rh, bp, cc, s->tx()))
		s->nf(1);
}

void SessionManager::recv(LoopProbeHandler* lph) 
{
	Network* n = lph->net();	
	n_long src;
	u_char* bp = pktbuf_;
	int cc = n->recv(bp, 2 * RTP_MTU, src);

	if (cc <= 0)
		return;
	
	LoopProbe::loopProbe* p = (LoopProbe::loopProbe*)bp;
	
	/* Random packet */
	if (p->magic != htonl(LOOPPROBE_MAGIC))
		return;

	n_long srcid = ntohl(p->srcid);
	n_long nhops = ntohl(p->hops);
	n_long* hops = (n_long*)(p + 1);
	if (srcid == A_.srcid() || srcid == B_.srcid()) {
		char buf[1024];
		*buf = '\0';
		int i;
		strcat(buf, intoa(LookupLocalAddr()));
		strcat(buf, " ");
		for (i = 0; i < nhops; i++) {
			strcat(buf, intoa(*hops++));
			strcat(buf, " ");
		}
		strcat(buf, intoa(LookupLocalAddr()));
		Tcl::instance().evalf("loopdetect \"%s\"", buf);
		/* Our probe.  A loop is present.  For now - exit. */
		adios();
	} else {
		/* If it's not our probe - forward it and add our host name */

		hops += nhops;
		*hops = LookupLocalAddr();
		p->hops = htonl(nhops + 1);

		LoopProbeHandler* h = (lph->sess() == &A_) ? B_.lph() : 
							     A_.lph();
		for (; h != 0; h = (LoopProbeHandler*)h->nexthandler)
			h->send(bp, cc + sizeof(n_long));
	}

}

void SessionManager::parse_sr(rtcphdr* rh, int, u_char*, Source* ps,
			      n_long addr)
{
	rtcp_sr* sr = (rtcp_sr*)(rh + 1);
	Source* s;
	u_int32_t ssrc = rh->rh_ssrc;
	if (ps->srcid() != ssrc)
		s = gwSourceManager::instance().lookup(ssrc, ssrc, addr);
	else
		s = ps;
#ifdef notdef
	s->touch_ctrl();
	s->ctrlntp(ntohl(sr->sr_ntp.upper) & 0xffff0000 |
		   ntohl(sr->sr_ntp.lower) >> 16);
	s->lctrlntp(ntptime());
#endif
	/* 
	 * If we receive an SR from the the source then we must 
	 * modify the report to reflect the output values.  
	 */
	if (s->handler() != 0) {
		gwSource* vgws = (gwSource*)s;
		sr->sr_np = htonl(vgws->opkts());
		sr->sr_nb = htonl(vgws->obits());
	}
}

void SessionManager::parse_rr(rtcphdr* rh, int, u_char*, Source* ps,
			      n_long addr)
{
	rtcp_rr* rr = (rtcp_rr*)(rh + 1);
	Source* s;
	u_int32_t ssrc = rh->rh_ssrc;
	if (ps->srcid() != ssrc)
		s = gwSourceManager::instance().lookup(ssrc, ssrc, addr);
	else
		s = ps;
#ifdef notdef
	s->touch_ctrl();
#endif
	/* 
	 * If we receive an RR from the destination then we must
	 * modify the report to reflect the input values.
	 */
	if (s->handler() != 0) {
                int received = s->np() - s->snp();
                s->snp(s->np());
                int expected = s->ns() - s->sns();
                s->sns(s->ns());
                u_int32_t v;
                int lost = expected - received;
                if (lost <= 0)
                        v = 0;
                else
                        /* expected != 0 if lost > 0 */
                        v = ((lost << 8) / expected) << 24;
                /* XXX should saturate on over/underflow */
                v |= (s->ns() - s->np()) & 0xffffff;
                rr->rr_loss = htonl(v);
                rr->rr_ehsr = htonl(s->ehs());
		rr->rr_dv = (s->handler() != 0) ? s->handler()->delvar() : 0;
	}
}

int SessionManager::sdesbody(n_long* p, u_char* ep, Source* ps, n_long addr)
{
	Source* s;
	n_long id = *p;
	if (ps->srcid() != id)
		s = gwSourceManager::instance().lookup(id, id, addr);
	else
		s = ps;
	u_char* cp = (u_char*)(p + 1);

	while (cp < ep) {
		char buf[256];

		u_int type = cp[0];
		if (type == 0) {
			/* end of chunk */
			return (((cp - (u_char*)p) >> 2) + 1);
		}
		u_int len = cp[1];
		u_char* eopt = cp + len + 2;
		if (eopt > ep)
			return (0);

		if (type >= RTCP_SDES_MIN && type <= RTCP_SDES_MAX) {
			memcpy(buf, (char*)&cp[2], len);
			buf[len] = 0;
			s->sdes(type, buf);
		} else
			/*XXX*/;
 
		cp = eopt;
	}
	return (0);
}

void SessionManager::parse_sdes(rtcphdr* rh, int flags, u_char* ep, Source* ps,
				n_long addr)
{
	int cnt = flags >> 8 & 0x1f;
	n_long* p = (n_long*)&rh->rh_ssrc;
	while (--cnt >= 0) {
		int n = sdesbody(p, ep, ps, addr);
		if (n == 0)
			break;
		p += n;
	}
	if (cnt >= 0)
		ps->badsdes(1);
}

void SessionManager::parse_bye(rtcphdr* rh, int flags, u_char* ep, Source* ps)
{
	int cnt = flags >> 8 & 0x1f;
	n_long* p = (n_long*)&rh->rh_ssrc;

	while (--cnt >= 0) {
		if (p >= (n_long*)ep) {
			ps->badbye(1);
			return;
		}
		Source* s;
		if (ps->srcid() != rh->rh_ssrc)
			s = gwSourceManager::instance().consult(*p);
		else
			s = ps;
		if (s != 0)
			s->lts_done(unixtime());
		++p;
	}
}

/*
 * Receive an rtcp packet (from the control port).
 */
void SessionManager::recv(CtrlHandler* ch)
{
	Network* net = ch->net();
	n_long src;
	int cc = net->recv(pktbuf_, 2 * RTP_MTU, src);
	if (cc <= 0)
		return;

	/* 
	 * Determine if we can accept the packet. 
	 */
	if (A_.ch() == 0 || B_.ch() == 0)
		return;

	rtcphdr* rh = (rtcphdr*)pktbuf_;
	if (cc < sizeof(*rh)) {
		++nrunt_;
		return;
	}
	/*
	 * try to filter out junk: first thing in packet must be
	 * sr, rr or bye & version number must be correct.
	 */
	switch(ntohs(rh->rh_flags) & 0xc0ff) {
	case RTP_VERSION << 14 | RTCP_PT_SR:
	case RTP_VERSION << 14 | RTCP_PT_RR:
	case RTP_VERSION << 14 | RTCP_PT_BYE:
		break;
	default:
		++badversion_;
		return;
	}
	/*
	 * at this point we think the packet's valid.  Update our average
	 * size estimator.  Also, there's valid ssrc so charge errors to it
	 */
	rtcp_avg_size_ += RTCP_SIZE_GAIN * (double(cc + 28) - rtcp_avg_size_);
	n_long addr = src;
	SourceManager& sm = gwSourceManager::instance();
	u_int32_t ssrc = rh->rh_ssrc;
	Source* ps = sm.lookup(ssrc, ssrc, addr);
	u_char* epack = (u_char*)rh + cc;
	while ((u_char*)rh < epack) {
		u_int len = (ntohs(rh->rh_len) << 2) + 4;
		u_char* ep = (u_char*)rh + len;
		if (ep > epack) {
			ps->badsesslen(1);
			return;
		}
		u_int flags = ntohs(rh->rh_flags);
		if (flags >> 14 != RTP_VERSION) {
			ps->badsessver(1);
			return;
		}
		switch (flags & 0xff) {

		case RTCP_PT_SR:
			parse_sr(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_RR:
			parse_rr(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_SDES:
			parse_sdes(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_BYE:
			parse_bye(rh, flags, ep, ps);
			break;

		default:
			ps->badsessopt(1);
			break;
		}
		rh = (rtcphdr*)ep;
	}
		
	CtrlHandler* h = (ch->sess() == &A_) ? B_.ch() : A_.ch();
	for (; h != 0; h = (CtrlHandler*)h->nexthandler)
		h->send(pktbuf_, cc);
}

int SessionManager::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);

	const char* cmd = argv[1];
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(cmd, "init") == 0) {
			A_.init();
			B_.init();
			return (TCL_OK);
		}
		if (strcmp(cmd, "A-session") == 0) {
			tcl.resultf("%s", A_.name());
			return (TCL_OK);
		}
		if (strcmp(cmd, "B-session") == 0) {
			tcl.resultf("%s", B_.name());
			return (TCL_OK);
		}
	}
}

Session::Session(const char* name, SessionManager *sm)
	:sname_(name), sm_(sm), rc_(1), dh_(0), ch_(0), lph_(0), datanets_(0),
	 color_(1)
{
	srcid_ = alloc_srcid(LookupLocalAddr());
}

Session::~Session()
{
	/* Delete all handlers */
}

void Session::init(void)
{
	DataHandler* p;
	for (p = dh_; p != 0; p = p->nexthandler)
		p->unlink();
	for (p = ch_; p != 0; p = p->nexthandler)
		p->unlink();
	for (p = lph_; p != 0; p = p->nexthandler)
		p->unlink();
	*ofmt_ = '\0';
}

int Session::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);
	
	const char* cmd = argv[1];
	if (argc == 2) {
		if (strcmp(cmd, "ofmt") == 0) {
			Tcl::instance().result(ofmt_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "rate-control") == 0) {
			Tcl::instance().resultf("%d", rc_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "probe-init") == 0) {
			lp_.init(lph_, srcid_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "session-name") == 0) {
			Tcl::instance().resultf("%s", sname_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "color") == 0) {
			Tcl::instance().resultf("%d", color_);
			return (TCL_OK);
		}
#ifdef notdef
		if (strcmp(cmd, "data-net") == 0) {
			Tcl::instance().resultf("%s", dh_.net()->name());
			return (TCL_OK);
		}
#endif
	}
	if (argc == 3) {
		if (strcmp(cmd, "ofmt") == 0) {
			strcpy(ofmt_, argv[2]);
			return (TCL_OK);
		} 
		if (strcmp(cmd, "rate-control") == 0) {
			rc_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(cmd, "color") == 0) {
			color_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(cmd, "data-net") == 0) {
			DataHandler* h = new DataHandler(*sm_, this);
			Network* n = (Network*)TclObject::lookup(argv[2]);
			h->net(n);
			n->nextnet = datanets_;
			datanets_ = n;
			h->nexthandler = dh_;
			dh_ = h;
			return (TCL_OK);
		}
		if (strcmp(cmd, "ctrl-net") == 0) {
			CtrlHandler* h = new CtrlHandler(*sm_, this);
			h->net((Network*)TclObject::lookup(argv[2]));
			h->nexthandler = ch_;
			ch_ = h;
			return (TCL_OK);
		}
		if (strcmp(cmd, "loop-probe-net") == 0) {
			LoopProbeHandler* h = new LoopProbeHandler(*sm_, this);
			h->net((Network*)TclObject::lookup(argv[2]));
			h->nexthandler = lph_;
			lph_ = h;
			return (TCL_OK);
		}
	} 
	return (TclObject::command(argc, argv));
}

u_int32_t Session::alloc_srcid(u_int32_t addr) const
{
	timeval tv;
	::gettimeofday(&tv, 0);
	u_int32_t srcid = u_int32_t(tv.tv_sec + tv.tv_usec);
	srcid += (u_int32_t)getuid();
	srcid += (u_int32_t)getpid();
	srcid += addr;
	return (srcid);
}

void LoopProbe::init(LoopProbeHandler* l, u_int32_t srcid) 
{
	lplist_ = l;
	loopProbe* p = (loopProbe*)buf_;
	p->magic = htonl(LOOPPROBE_MAGIC);
	p->srcid = htonl(srcid);
	p->hops = 0;
	
	timeout();
}
