/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/RCS/session.h,v 1.16 1996/05/15 03:28:12 elan Exp $ (LBL)
 */

#ifndef vgw_session_h
#define vgw_session_h

extern "C" {
#include <tk.h>
}
#include "inet.h"
#include "rtp.h"
#include "iohandler.h"
#include "timer.h"
#include "Tcl.h"
#include "net.h"

class Source;
class Session;
class SessionManager;
class LoopProbeHandler;

class DataHandler : public IOHandler {
 public:
	DataHandler* nexthandler;
 public:
	inline DataHandler(SessionManager& sm, Session* s) : sm_(sm), net_(0),
		sess_(s) {}
	virtual void dispatch(int);
	inline Network* net() const { return (net_); }
	inline Session* sess() const { return (sess_); }
	inline void net(Network* n) {
		unlink();
		link(n->rchannel(), TK_READABLE);
		net_ = n;
	}
	inline int recv(u_char* bp, int len, u_int32_t& addr) {
		return (net_->recv(bp, len, addr));
	}
	inline void send(u_char* bp, int len) {
		net_->send(bp, len);
	}
protected:
	SessionManager& sm_;
	Network* net_;
	Session* sess_;
};

class CtrlHandler : public DataHandler {
 public:
	inline CtrlHandler(SessionManager& sm, Session* s) 
		: DataHandler(sm, s) {}
	virtual void dispatch(int);
};

class LoopProbeHandler : public DataHandler {
 public:
	inline LoopProbeHandler(SessionManager& sm, Session* s) 
		: DataHandler(sm, s) {}
	virtual void dispatch(int);
};

#define LOOPPROBE_INTERVAL 10000	/* 10 sec */
class LoopProbe : public Timer {
public:
	LoopProbe() : lplist_(0) {}
	struct loopProbe {
#define LOOPPROBE_MAGIC	0xdeadbeef		
		u_int32_t magic;
		u_int32_t srcid;
		u_int32_t hops;
	};
	void init(LoopProbeHandler* l, u_int32_t srcid);
	virtual void timeout(void) {
		LoopProbeHandler* p;
		for (p = lplist_; p != 0; 
		     p = (LoopProbeHandler*)p->nexthandler)
			p->send(buf_, sizeof(loopProbe));
		msched(LOOPPROBE_INTERVAL);
	}
private:
	u_char buf_[sizeof(loopProbe)];
	LoopProbeHandler* lplist_;
};

class SessionManager;

class Session : public TclObject {
 public:
	Session(const char* name, SessionManager* sm);
	~Session();
	inline DataHandler* dh() { return (dh_); }
	inline CtrlHandler* ch() { return (ch_); }
	inline LoopProbeHandler* lph() { return (lph_); }
	inline const char* sname() { return sname_; }
	inline char* ofmt() { return (ofmt_); }
	inline Network* datanets() { return (datanets_); }
	inline u_int32_t srcid() const { return (srcid_); }
	void init();
 protected:
	virtual int command(int, const char*const*);
	u_int32_t alloc_srcid(u_int32_t addr) const;
	DataHandler* dh_;
	CtrlHandler* ch_;
	LoopProbeHandler* lph_;
	LoopProbe lp_;
	const char* sname_;
	SessionManager* sm_;
	u_int32_t srcid_;
	int rc_;
	char ofmt_[32];
	Network* datanets_;
	int color_;
};
		

class SessionManager : public TclObject {
 public:
	SessionManager();
	virtual void recv(DataHandler*);
	virtual void recv(CtrlHandler*);
	virtual void recv(LoopProbeHandler*);
	virtual int command(int, const char*const*);
 protected:
	u_int32_t ntptime(timeval t);
	u_int32_t ntptime();

	void parse_sr(rtcphdr* rh, int flags, u_char* ep, Source* ps, 
		      u_int32_t addr);
	void parse_rr(rtcphdr* rh, int flags, u_char* ep, Source* ps, 
		      u_int32_t addr);
	int sdesbody(u_int32_t* p, u_char* ep, Source* ps, u_int32_t addr);
	void parse_sdes(rtcphdr* rh, int flags, u_char* ep, Source* ps, 
			u_int32_t addr);
	void parse_bye(rtcphdr* rh, int flags, u_char* ep, Source* ps);

	int parseopts(const u_char* bp, int cc, u_int32_t addr) const;
	int recv(int fd, sockaddr_in& src);
	int ckid(const char*, int len);
	virtual int check_format(int fmt) = 0;

	/*XXX cleanup*/
	u_int badversion_;
	u_int badoptions_;
	u_int badext_;
	u_int badfmt_;
	u_int nrunt_;

	u_int32_t last_np_;

	double rtcp_avg_size_;	/* (estimated) average size of rtcp packets */
	u_char pktbuf_[2 * RTP_MTU];
	
	Session A_;
	Session B_;
};


class AudioSessionManager : public SessionManager {
 protected:
	virtual int check_format(int fmt);
};

class VideoSessionManager : public SessionManager {
 protected:
	virtual int check_format(int fmt);
};

#endif
