/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/engines/RCS/source-gw.cc,v 1.30 1996/05/23 22:30:55 elan Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

#include "net.h"
#include "transcoder.h"
#include "inet.h"
#include "Tcl.h"
#include "transmitter.h"
#include "source-gw.h"

vgwSource::vgwSource(n_long srcid, u_int32_t ssrc, u_int32_t addr)
	 :gwSource(srcid, ssrc, addr)
{}

agwSource::agwSource(n_long srcid, u_int32_t ssrc, u_int32_t addr)
	 :gwSource(srcid, ssrc, addr)
{}

gwSource::gwSource(n_long srcid, u_int32_t ssrc, u_int32_t addr)
	 :Source(srcid, ssrc, addr)
{
	osess_ = "";
}

PacketHandler* gwSource::activate(int format, Network* net, const char* osess)
{
	format_ = format;
	osess_ = osess;

	tx_.ssrc(ssrc_);
	tx_.net(net);
	
	Tcl::instance().evalf("activate %s", TclObject::name());
	if (handler_ == 0)
		/* tcl should install a handler */
		abort();
	return (handler_);
}

int gwSource::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);

	Tcl& tcl = Tcl::instance();

	const char* cmd = argv[1];
	if (argc == 2) {
		if (strcmp(argv[1], "obits") == 0) {
			tcl.resultf("%d", 8 * tx_.nb());
			return (TCL_OK);
		}
		if (strcmp(argv[1], "opkts") == 0) {
			tcl.resultf("%d", tx_.np());
			return (TCL_OK);
		}
		if (strcmp(argv[1], "ofrms") == 0) {
			tcl.resultf("%d", tx_.nf());
			return (TCL_OK);
		}
		if (strcmp(cmd, "bps") == 0) {
			tcl.resultf("%d", tx_.bps());
			return (TCL_OK);
		}
		if (strcmp(cmd, "ifmt") == 0) {
 			tcl.resultf("%d", format_);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "osess") == 0) {
			tcl.resultf("%s", osess_);
			return (TCL_OK);
		}
		if (strcmp(cmd, "transcoder") == 0) {
			Transcoder* t = (Transcoder*)handler_;
			if (t != 0)
				tcl.resultf("%s", t->name());
			else 
				tcl.result("");
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(cmd, "bps") == 0) {
			u_int b = atoi(argv[2]);
			tx_.bps(b);
			return (TCL_OK);
		}
		if (strcmp(cmd, "transcoder") == 0) {
			delete (Transcoder*)handler_;
			handler_ = (Transcoder*)TclObject::lookup(argv[2]);
			return (TCL_OK);
		}
	}

	return (Source::command(argc, argv));
}

int vgwSource::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);

	Tcl& tcl = Tcl::instance();

	const char* cmd = argv[1];
	if (argc == 2) {
		if (strcmp(cmd, "width") == 0) {
			VideoTranscoder* t = (VideoTranscoder*)handler_;
			tcl.resultf("%d", t->width());
			return (TCL_OK);
		}
		if (strcmp(cmd, "height") == 0) {
			VideoTranscoder* t = (VideoTranscoder*)handler_;
			tcl.resultf("%d", t->height());
			return (TCL_OK);
		}
		if (strcmp(cmd, "decimation") == 0) {
			VideoTranscoder* t = (VideoTranscoder*)handler_;
			tcl.resultf("%d", t->decimation());
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(cmd, "color") == 0) {
			VideoTranscoder* t = (VideoTranscoder*)handler_;
			t->color(atoi(argv[2]));
			return (TCL_OK);
		}
	}
	return (gwSource::command(argc, argv));
}

int agwSource::command(int argc, const char*const* argv)
{
	if (argc < 2)
		return (TCL_ERROR);

	Tcl& tcl = Tcl::instance();

	return (gwSource::command(argc, argv));
}

gwSourceManager* gwSourceManager::instance_;

gwSourceManager::gwSourceManager()
{
	if (instance_ != 0)
		abort();
	instance_ = this;
}

Source* agwSourceManager::create(u_int32_t srcid, u_int32_t ssrc, 
				 u_int32_t addr)
{
	return (new agwSource(srcid, ssrc, addr));
}

Source* vgwSourceManager::create(u_int32_t srcid, u_int32_t ssrc, 
				 u_int32_t addr)
{
	return (new vgwSource(srcid, ssrc, addr));
}

static class gwSourceManagerMatcher : public Matcher {
public:
	gwSourceManagerMatcher() : Matcher("source-manager") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "audio") == 0)
			return (new agwSourceManager());
		else if (strcmp(type, "video") == 0) 
			return (new vgwSourceManager());
		return (0);
	}
} gwsource_manager_matcher;


			
				
			
	
