/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/engines/RCS/source-gw.h,v 1.13 1996/11/12 19:06:25 elan Exp $ (LBL)
 */

#ifndef vgw_source_vgw_h
#define vgw_source_vgw_h

#include "config.h"
#include "source.h"
#include "transmitter.h"

class Network;

class gwSource : public Source {
 public:
	gwSource(u_int32_t srcid, u_int32_t ssrc, u_int32_t addr);
	virtual int command(int argc, const char*const* argv);
	PacketHandler* activate(int, Network*, const char* osess);

	inline u_int32_t obits() const { return (8 * tx_.nb()); }
	inline u_int32_t ofrms() const { return (tx_.nf()); }
	inline u_int32_t opkts() const { return (tx_.np()); }
	inline Transmitter& tx() { return (tx_); }

	inline osess(const char* s) { osess_ = s; }
 protected:
	Transmitter tx_;
	const char* osess_;
};

class vgwSource : public gwSource {
 public:
	vgwSource(u_int32_t srcid, u_int32_t ssrc, u_int32_t addr);
	virtual int command(int argc, const char*const* argv);
};

class agwSource : public gwSource {
 public:
	agwSource(u_int32_t srcid, u_int32_t ssrc, u_int32_t addr);
	virtual int command(int argc, const char*const* argv);
};

class gwSourceManager : public SourceManager {
 public:
	gwSourceManager();
	static inline gwSourceManager& instance() { return (*instance_); }
	virtual Source* create(u_int32_t srcid, u_int32_t ssrc, 
			       u_int32_t addr) = 0;
 private:
	static gwSourceManager* instance_;
};

class vgwSourceManager : public gwSourceManager {
 public:
	virtual Source* create(u_int32_t srcid, u_int32_t ssrc,
			       u_int32_t addr);
};

class agwSourceManager : public gwSourceManager {
 public:
	virtual Source* create(u_int32_t srcid, u_int32_t ssrc,
			       u_int32_t addr);
};

#endif
