#ifndef vgw_transcoder_h
#define vgw_transcoder_h

#include <sys/types.h>
#include <sys/time.h>
#include "config.h"
#include "inet.h"
#include "source.h"
#include "encoder.h"
#include "Tcl.h"

class Transcoder;
class Network;
class Transmitter;

class Transcoder : public PacketHandler {
 public:
	virtual int command(int argc, const char*const* argv);
	virtual int recv(const rtphdr*, const u_char* pkt, int len, 
			 Transmitter& tx) = 0;
	inline double gettimeofday() {
		timeval tv;
		::gettimeofday(&tv, 0);
		return (tv.tv_sec * 1e6 + tv.tv_usec);
	}
 protected:
	Transcoder(int hdrlen);
	virtual const char* media_type(void) = 0;

	double fc_;
	int forward_;
	char tname_[24];
};

class VideoTranscoder : public Transcoder {
 public:
	virtual int command(int argc, const char*const* argv);
	inline int width() const { return (inw_); }
	inline int height() const { return (inh_); }
	inline int decimation() const { return (decimation_); }
	inline void color(int v) { 
		if (encoder_ != 0)
			encoder_->color(v);
		send_all();
	}
 protected:
	VideoTranscoder(int hdrlen);
	~VideoTranscoder();
	VideoEncoder* encoder_;
	virtual const char* media_type(void) { return ("video"); }
	void crinit(int, int);
	void age_blocks(void);
	void send_all(void);
	int decimation_;	/* 411 or 422 */
	int inw_;		/* input image width */
	int inh_;		/* input image height */
	int outw_;		/* output image width */
	int outh_;		/* output image height */

	/* CR stuff */
	int scan_;
	int nblk_;
	u_char* crvec_;
	int rover_;

};

#define PLAYO_FILTER 5
#define VAR_FILTER 5
#define VAR_MULT 2

class AudioTranscoder : public Transcoder {
    protected:
	AudioTranscoder(int hdrlen);
    public:
	virtual ~AudioTranscoder();
	virtual int command(int argc, const char*const* argv);
	inline u_int32_t delvar() const { return (delvar_); }
	inline void delvar(u_int32_t v) { delvar_ = v; }
    protected:
	virtual const char* media_type(void) { return ("audio"); }
	int adapt(const rtphdr*, u_int32_t sclock);
	inline u_int32_t playout() const { return (playout_ >> PLAYO_FILTER); }
	/*
	 * Playback point estimation state.
	 */
	int32_t var_;		// variance in this host's interarrival time
	u_int32_t hostoffset_;	// timestamp offset this host's clock
	u_int32_t predicted_drop_;	// ts of last drop + blocksize
	u_int32_t playout_;	// playout delay (in media units)
	u_int32_t lastrecv_;	// local time (in media time) of last recv
	int lecture_mode_;	/*XXX*/
	int framesize_;		/*XXX*/
	int block_size_;
	int maxdel_;

	AudioEncoder* encoder_;

	int delvar_;
};

class NullTranscoder : public Transcoder {
 public:	
	NullTranscoder() : Transcoder(0) {}
	int recv(const rtphdr*, const u_char* pkt, int len, Transmitter& tx);
 protected:
	virtual const char* media_type(void) { return ("pass-thru"); }
 private:
	u_char pktbuf_[8192];	/* XXX */
};

#endif /* vgw_transcoder_h */
