#ifndef vgw_encoder_h261_h
#define vgw_encoder_h261_h

#include <sys/types.h>
#include "encoder.h"
#include "config.h"
#include "transmitter.h"

class H261Encoder : public VideoEncoder {
public:
	virtual int consume(const VideoFrame*, Transmitter&);
	virtual int encode(n_long ts, const u_char* frame, 
			   const u_int8_t *crvec, Transmitter&);
	virtual int framesize(int inw, int inh, int& outw, int& outh) const;
protected:
	H261Encoder();
#ifdef notdef
	int command(int argc, const char*const* argv);
#endif
	int flush(Transmitter::pktbuf* pb, int nbit, Transmitter::pktbuf* npb,
		  Transmitter&);
	void setq(int q);
	void size(int w, int h, int dec);

	virtual void encode_mb(int mba, const u_char* frm, int blkno, 
			       int uvblkno, int boff, int how) = 0;
	void encode_blk(const short* blk, const char* lm);
 	char* make_level_map(int q, u_int fthresh);
	void setquantizers(int lq, int mq, int hq);
	
	/* bit buffer */
	u_int bb_;
	u_int nbb_;

	u_char* bs_;
	u_int bc_;
	int sbit_;

	int mquant_;
	int ngob_;
	int mba_;

	int cif_;		/* 1 for CIF, 0 for QCIF */
	int bstride_;
	int lstride_;
	int cstride_;

	int quant_required_;
	const u_char* lscale_;
	const u_char* cscale_;
	float lqt_[64];		/* low quality quantizer */
	float mqt_[64];		/* medium quality quantizer */
	float hqt_[64];		/* high quality quantizer */
	
	int lq_;
	int mq_;
	int hq_;
	
	const char* llm_[32];
	const char* clm_[32];

	int blkno_[12];
	int uvblkno_[12];

	int decimation_;

	int width_;
	int height_;
	int framesize_;
};

class H261DCTEncoder : public H261Encoder {
 private:
	virtual void encode_mb(int mba, const u_char* frm, int blkno, int 
			       uvblkno, int boff, int how);
 public:
	H261DCTEncoder();
};

class H261PixelEncoder : public H261Encoder {
 private:
	virtual void encode_mb(int mba, const u_char* frm, int blkno, 
			       int uvblkno, int boff, int how);
 public:
	H261PixelEncoder();
};

#endif /* vgw_encoder_h261_h */
