#
# Copyright (c) 1993-1995 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#	This product includes software developed by the Computer Systems
#	Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: /home/barad-dur/b/users/elan/src/rtpgw/ui/RCS/help.tcl,v 1.8 1996/08/10 01:30:30 elan Exp $ (UC Berkeley)
#

set helpno 0
proc helpitem { w text } {
	global helpno
#    	set f [option get . helpFont RtpgwUI]
	set h $w.h$helpno
	incr helpno
	frame $h 
	canvas $h.bullet -width 12 -height 12 -highlightthickness 0
	$h.bullet create oval 6 3 12 9 -fill black
#	message $h.msg -justify left -anchor w -font $f -width 460 -text $text
	message $h.msg -justify left -anchor w -width 460 -text $text
	pack $h.bullet -side left -anchor ne -pady 5
	pack $h.msg -side left -expand 1 -fill x -anchor nw
	pack $h -expand 1 -fill both
}

proc build.help { } {
	set w .help
	if [winfo exists $w] { return }
	toplevel $w
	wm title $w "rtpgw menu"
	wm withdraw $w
	bind $w <Enter> "focus $w"

	set p [frame $w.frame -borderwidth 2 -relief sunken]
	helpitem $p "Click on the \"Menu\" button to bring up the control menu."
	helpitem $p "Session Rate Control:  To enable rate control for the session, activate the approriate \"Rate Control\" button.  In general, the high bandwidth link (default A) should have rate control turned off and the lower bandwidth link (default B) session should have rate control turned on."
        helpitem $p "Per Session Adjustments: For a session with rate control, adjust the output bandwidth and output format appropriately.  If rate control is not set for a session, packets destined for that session will be simply be forwarded by the gateway."
	helpitem $p "Voice Switched Bandwidth Allocation: Activating the \"Switching\" button in a rate controlled session enables voice switched bandwidth allocation.  See the manual page for further details."
	helpitem $p "Source Box: Each source is displayed as shown below.  Each source has 3 lines of information.  The top line contains the source session (A or B) and name.  The middle line contains the address information of the source (RTP cname), the input format from the source, and the transcoder in place for this source.  Finally, the third line shows the input and output frame rates and bandwidths.  In addition to the above information, the source box contains a bandwidth allocation slider that enables allocation to each stream of a fraction of the total output session bandwidth.  Note that the slider will prevent allocating over 100% of the output session bandwidth among all streams outputting to the given session.  Sources on a non-rate controlled session will have their bandwidth allocation slider and frame rate values disabled."
	set s [build.helpsrc $p]
	pack $s -pady 5
	helpitem $p "Click on the \"Engine\" button to see the information on the gateway engine."
	helpitem $p "Exiting the user interface kills the engine as well."
	helpitem $p "Bugs and suggestions to rtpgw@daedalus.CS.Berkeley.EDU. Thanks."

	pack $w.frame -expand 1 -fill both -pady 5 -padx 10
	button $w.dismiss -text " Dismiss " -borderwidth 2 -relief raised \
		-command "wm withdraw $w" -font [mediumfont] 
	pack $w.dismiss -pady 5
}

proc build.helpsrc { inw } {
	global config

	set f [smallfont]

	set w [frame $inw.help]

	# for compatibility with mono-mode 
	mkSrcFrame $w.l 
	frame $w.l.name 
	frame $w.l.info 
	frame $w.l.rates

	mkSrcFrame $w.r
	frame $w.r.slider
	
	global srcinfo
	label $w.l.name.srcspec -text "Elan Amir (UC Berkeley)" -font $f
	label $w.l.name.sname -text "(A)" -font $f
	pack  $w.l.name.sname -side left
	pack $w.l.name.srcspec -expand 1 -fill x -side left
	pack $w.l.name -side top

	label $w.l.info.srcspec -text "elan@128.32.33.58/jpeg" -font $f
	label $w.l.info.tname -text "jpeg->h261" -font $f 
	pack $w.l.info.srcspec $w.l.info.tname -side left -expand 1
	pack $w.l.info -side top -fill x
	
	label $w.l.rates.ifps -text "29 f/s" -font $f 	-width 8 -height 1
	label $w.l.rates.ibps -text "1.2 Mb/s" -font $f -width 8 -height 1
	label $w.l.rates.arrow -text "  -->" -font $f -width 5 -height 1
	label $w.l.rates.ofps -text "27 f/s" -font $f -width 8 -height 1
	label $w.l.rates.obps -text "282 kb/s" -font $f -width 8 -height 1

	pack $w.l.rates.ifps $w.l.rates.ibps $w.l.rates.arrow $w.l.rates.ofps \
		$w.l.rates.obps -side left
	pack $w.l.name -side top -fill both -expand 1
	pack $w.l.info -side top
	pack $w.l.rates -side top -fill both -expand 1 

	mkHelpSrcBWController $w.r

	pack $w.l -side left
	pack $w.r -side left -fill both -expand 1
	
	return $w
}

proc mkHelpSrcBWController { w } {
	global config 
	set f [smallfont]

	scale $w.slider.scale -orient horizontal -from 0 -to 100 \
			-showvalue 0 -font $f -width 10 \
			-highlightthickness 0 
	frame $w.slider.info
	label $w.slider.info.name -text " %bw" -font $f -width 3
    	$w.slider.scale set 56
	label $w.slider.info.val -text "56.0" -font $f -width 4
	label $w.slider.info.maxbpsval -font $f -width 8 -text "286 kb/s"
	pack $w.slider.info.val $w.slider.info.name \
		$w.slider.info.maxbpsval -side left -expand 1 -fill x
	pack $w.slider.scale $w.slider.info -side bottom -expand 1
	pack $w.slider -side top -fill both -expand 1
    	$w.slider.scale configure -state disabled
}
