/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) $Header: /home/barad-dur/b/users/elan/src/vgw/ui/RCS/main.cc,v 1.30 1996/11/11 21:17:14 elan Exp $ (LBL)";

#include <osfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/stat.h>
#ifdef sgi
#include <getopt.h>
#endif
#include <X11/Xlib.h>

extern "C" {
#include <tk.h>
}

#if defined(sun) && !defined(__svr4__)
extern "C" int gettimeofday(struct timeval*, struct timezone*);
extern "C" int nice(int incr);
#endif
#if defined(sun) || defined(__osf__)
extern "C" int gethostname(char* name, int len);
#endif

#include "inet.h"
#include "Tcl.h"

extern "C" char version[];
extern "C" char tcl_rtpgw_ui[];
extern "C" char tcl_tk[];
extern "C" char tcl_local[];

static void
usage()
{
#ifdef INFOPAD
	fprintf(stderr, "\
Usage: rtpgw_ui -T [audio|video] [-I channel] [-S] [-u script] [-U interval] [-v] [-x engine] Asession Bsession\n\
       or\n\
       rtpgw_ui -T [audio|video] [-I channel] [-S] [-u script] [-U interval] [-v] [-x engine] -p padid Bsession\n (session: addr/port[/ttl])\n");
#else
	fprintf(stderr, "\
Usage: rtpgw_ui -T [audio|video] [-I channel] [-K keyA:keyB] [-S] [-u script] [-U interval]\n\
               [-v] [-x engine] Asession Bsession\n\
       (session: addr/port[/ttl])\n");
#endif
	exit(1);
}

pid_t engine_pid = -1;

SIGRET
adios(...)
{
	if (engine_pid > 0)
		kill(engine_pid, SIGHUP);
	exit(0);
}

static class EnginePidCommand : public TclObject {
public:
	EnginePidCommand() : TclObject("engine_pid") {}
	int command(int argc, const char*const* argv) {
		if (argc != 2) 
			return (TCL_ERROR);
		engine_pid = atoi(argv[1]);
		return(TCL_OK);
	}
} cmd_engine_pid;

static class AdiosCommand : public TclObject {
public:
	AdiosCommand() : TclObject("adios") {}
	int command(int argc, const char*const* argv) {
		adios();
		/*NOTREACHED*/
		return (0);
	}
} cmd_adios;
 
static class VersionCommand : public TclObject {
public:
	VersionCommand() : TclObject("version") {}
	int command(int argc, const char*const* argv) {
		Tcl::instance().result(version);
		return (TCL_OK);
	}
} cmd_version;

static class KillCommand : public TclObject {
public:
	KillCommand() : TclObject("kill") {}
	int command(int argc, const char*const* argv) {
		if (argc != 2) 
			return (TCL_ERROR);

		kill(atoi(argv[1]), SIGHUP);
		return (TCL_OK);
	}
} cmd_kill;

static class HaveFontCommand : public TclObject {
public:
	HaveFontCommand() : TclObject("havefont") {}
	int command(int argc, const char*const* argv) {
		Tcl& t = Tcl::instance();
		if (argc != 2)
			t.result("0");
		else {
			Tk_Window tk = t.tkmain();
			Tk_Uid uid = Tk_GetUid((char*)argv[1]);
			XFontStruct* p = Tk_GetFontStruct(t.interp(), tk, uid);
			t.result(p != 0 ? "1" : "0");
		}
		return (TCL_OK);
	}
} cmd_havefont;

extern "C" char *optarg;
extern "C" int optind;
extern "C" int opterr;

const char*
disparg(int argc, const char** argv, const char* optstr)
{
	const char* display = 0;
	int op;
	while ((op = getopt(argc, (char**)argv, (char*)optstr)) != -1) {
		if (op == 'd') {
			display = optarg;
			break;
		}
		else if (op == '?')
			usage();
	}
	optind = 1;
	return (display);
}

char*
parse_assignment(char* cp)
{
	cp = strchr(cp, '=');
	if (cp != 0) {
		*cp = 0;
		return (cp + 1);
	} else
		return ("true");
}

extern const char tcl_tkerror[];


extern "C" {
extern void TkCreateXEventSource();

int
TkPlatformInit(Tcl_Interp *interp)
{
        Tcl_SetVar(interp, "tk_library", ".", TCL_GLOBAL_ONLY);
	TkCreateXEventSource();
        return (TCL_OK);
}
}

int
main(int argc, const char** argv)
{
	signal(SIGHUP, adios);
	signal(SIGINT, adios);
	signal(SIGTERM, adios);
	signal(SIGCHLD, SIG_IGN);

	opterr = 0;
	const char* optstr = "A:I:K:p:n:ST:u:U:vX:x:";
	Tcl::init("rtpgwUI");
	Tcl& tcl = Tcl::instance();
	const char* display = disparg(argc, argv, optstr);

	tcl.evalf(display?
		  "set argv \"-name rtpgwUI -display %s\"" :
		  "set argv \"-name rtpgwUI\"",
		  display);
	Tk_Window tk = 0;
	if (Tk_Init(tcl.interp()) == TCL_OK)
		tk = Tk_MainWindow(tcl.interp());
	if (tk == 0) {
		fprintf(stderr, "rtpgw_ui: %s\n", tcl.result());
		exit(1);
	}
	tcl.tkmain(tk);

	tcl.evalc(tcl_tk);
	tcl.evalc(tcl_tkerror);
	tcl.evalc(tcl_local);

	tcl.add_default("updateInterval", "1000");
	tcl.add_default("sessionAtotalOutputBW", "128000");
	tcl.add_default("sessionBtotalOutputBW", "128000");
	tcl.add_default("sessionAspeakerBWpercent", "100");
	tcl.add_default("sessionBspeakerBWpercent", "100");
	tcl.add_default("sessionArateControl", "off");
	tcl.add_default("sessionBrateControl", "on");
	tcl.add_default("sessionAcolor", "1");
	tcl.add_default("sessionBcolor", "1");
	tcl.add_default("sessionAswitching", "off");
	tcl.add_default("sessionBswitching", "off");
	tcl.add_default("sessionAmaxBWScale", "5000000");
	tcl.add_default("sessionBmaxBWScale", "5000000");
	tcl.add_default("sessionBnetworkType", "ip");
	tcl.add_default("sessionAdefaultTTL", "4");
	tcl.add_default("sessionBdefaultTTL", "4");
	tcl.add_default("confBusChannel", "3");
	tcl.add_default("standalone", "no");
	tcl.add_default("engine", "");
	tcl.add_default("videoEngine", "vgw");
	tcl.add_default("audioEngine", "agw");
	tcl.add_default("tile", "8");
	tcl.add_default("proto", "vic");
	tcl.add_default("filterGain", "0.25");
	tcl.add_default("sessionAdefaultFormat", "h261");
#ifdef INFOPAD
	tcl.add_default("sessionBdefaultFormat", "ipadvq");
#else	
	tcl.add_default("sessionBdefaultFormat", "h261");
#endif
	const char* script = 0;
	int op;
	while ((op = getopt(argc, (char**)argv, (char*)optstr)) != -1) {
		switch (op) {
		default:
			usage();
		case 'A':
			tcl.add_option("proto", optarg);
			break;
		case 'I':
			tcl.add_option("confBusChannel", optarg);
			break;
		case 'K':
			tcl.add_option("sessionKeyString", optarg);
			break;
		case 'p':
#ifdef INFOPAD
			tcl.add_option("sessionBnetworkType", "ipad");
			tcl.add_option("padID", optarg);
#else
			fprintf(stderr, 
				"rtpgw_ui: No Berkeley Infopad support.\n");
			exit(1);
#endif /* INFOPAD */			
			break;
		case 'S':
			tcl.add_option("standalone", "yes");
			break;
		case 'T':
			tcl.add_option("mediaType", optarg);
			break;
		case 'u':
			script = optarg;
			break;
		case 'U':
			tcl.add_option("updateInterval", optarg);
			break;
		case 'v':
			printf("RtpgwUI version %s\n", version);
			exit(0);
			break;
		case 'x':
			tcl.add_option("engine", optarg);
			break;
		case 'X':
			{
				const char* value = parse_assignment(optarg);
				tcl.add_option(optarg, value);
			}
			break;
		}
	}

	int naddr;
	const char* onet = tcl.attr("sessionBnetworkType");
	if (strcmp(onet, "ip") == 0)
		naddr = 2;
	else if (strcmp(onet, "ipad") == 0) {
		naddr = 1;
	} else {
		fprintf(stderr, "rtpgw_ui: unknown network type %s.\n", onet);
		exit(1);
	}

	if (argc - optind != naddr)
	    usage();

	EmbeddedTcl::init();
	tcl.evalc(tcl_rtpgw_ui);

	if (script != 0)
		tcl.EvalFile(script);

	tcl.add_option("sessionAspec", argv[optind]);

	if (naddr == 2)
		tcl.add_option("sessionBspec", argv[optind + 1]);
	
	tcl.evalc("build.tk");

	Tk_MainLoop();

	adios();
}

