#
# Copyright (c) 1995 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: /home/barad-dur/b/users/elan/src/vgw/ui/RCS/mono.tcl,v 1.2 1995/12/12 22:54:08 elan Exp $ (UC Berkeley)
#
option add *background white 62
option add *foreground black 62
option add *highlightBackground black 62
option add *highlightForeground white 62
option add *activeBackground black 62
option add *activeForeground white 62
option add *selectBackground black 62
option add *selectForeground white 62
option add *disabledForeground white 62

proc build.bar w {
	global title
	frame $w.bar -highlightthickness 2 -borderwidth 2
	label $w.bar.title -text "VgwUI v[version]" -font [smallfont] \
		-relief flat -justify left
	button $w.bar.menu -text Menu -font [smallfont]  \
	    	-command "toggle_window .menu" -width 2
	button $w.bar.engine -text Engine -font [smallfont] \
		-command "toggle_window .engine" -width 3
	button $w.bar.help -text Help -font [smallfont]  \
		-command "toggle_window .help" -width 1
	button $w.bar.quit -text Quit -font [smallfont]  \
	    	-command "adios" -width 2
	pack $w.bar.title -side left -fill both -expand 1
	pack $w.bar.menu $w.bar.engine $w.bar.help $w.bar.quit -side left
}

proc build.menu {} {
	set w .menu
	toplevel $w
	wm title $w "vgw menu"
	wm withdraw $w
	bind $w <Enter> "focus $w"
	set f [smallfont]
	set med [mediumfont]
	
    	set b [frame $w.body]

	frame $b.info -highlightthickness 2 -borderwidth 1

	global config

    	label $b.info.title -text "Sessions"
	set A [frame $b.info.a -highlightthickness 1 -borderwidth 2]
    	build.session $A A
	set B [frame $b.info.b -highlightthickness 1 -borderwidth 2]
    	if { $config(onet) == "ipad" } {
		build.ipadsession $B B
	} else {
		build.session $B B
	}
	pack $b.info.title -side top -expand 1 -fill x
	pack $A $B -side top -anchor w -expand 1 -padx 4 -fill x -ipadx 2

	frame $b.buttons 	
	button $b.buttons.dismiss -text "Dismiss" -command "toggle_window $w"
	pack $b.buttons.dismiss -side left

	pack $b.info -side top -pady 5 -padx 4 -ipady 4 -expand 1 -fill x
	pack $b.buttons -side top -pady 5 -padx 4

	pack $b -side top
}

proc build.engine {} {
	set w .engine
	toplevel $w
	wm title $w "vgw engine"
	wm withdraw $w
	bind $w <Enter> "focus $w"
	set f [smallfont]
	set med [mediumfont]

	global config
	set sys [frame $w.sys -borderwidth 1 -highlightthickness 1]
	label $sys.t -text "Engine" -width 26

	frame $sys.status
	label $sys.status.title -text "Status:" -font $f
	label $sys.status.l -textvariable config(engine_status) -font $f 

	pack $sys.status.title $sys.status.l -side left -anchor w

	frame $sys.confbus
	label $sys.confbus.l -font $f \
		-text "Conference Bus Channel: $config(cb_channel)"

	pack $sys.confbus.l -side left
	pack $sys.t -side top
	pack $sys.status $sys.confbus -side top -anchor w 
	
	set ek [frame $sys.execkill]

	button $ek.exec -text Exec -font [smallfont] -width 1 \
		-command "execkill $ek exec"
	button $ek.kill -text Kill -font [smallfont] -width 1 \
		-command "execkill $ek kill"

	set config(pathnames,execbutton) $ek.exec
	set config(pathnames,killbutton) $ek.kill

	if { $config(vgw_engine) == "" } then {
		$ek.kill configure -state disabled
	} else {
		$ek.exec configure -state disabled
	}
	
	pack $ek.exec $ek.kill -side left -expand 1
	pack $ek -side top -fill x -pady 2

	frame $w.buttons 
	button $w.buttons.dismiss -text "Dismiss" -command "toggle_window $w"
	pack $w.buttons.dismiss -side top
	pack $sys $w.buttons -ipady 3 -padx 5 -pady 5 -anchor n
}

proc build.session { w name } {
	global config 
	set f [smallfont]

	frame $w.addr
	frame $w.bw

	label $w.addr.l \
	    -text "$name: $config($name,addr) port: $config($name,port) ttl: $config($name,ttl)" -font $f
	frame $w.addr.glue
	checkbutton $w.addr.cb -text "Rate Control" \
		-variable config($name,RC) -font $f -highlightthickness 1 \
		-command "set_RC $name"
	pack $w.addr.l -side left -anchor w
	pack $w.addr.glue -side left -expand 1 -fill x -anchor w
	pack $w.addr.cb -side right

	mkBottomLines $w $name
}

proc build.ipadsession { w name } {
	global config 
	set f [smallfont]

	frame $w.addr
	frame $w.bw

	label $w.addr.l -font $f \
		-text "$name: Infopad Session - Pad ID: $config(padid)" 
	frame $w.addr.glue
	checkbutton $w.addr.cb -text "Rate Control" \
		-variable config($name,RC) -font $f -highlightthickness 1 \
		-command "set_RC $name" 
	pack $w.addr.l -side left -anchor w
	pack $w.addr.glue -side left -expand 1 -fill x -anchor w
	pack $w.addr.cb -side right

	mkBottomLines $w $name
}

proc mkBottomLines {w name} {
	global config 
	set f [smallfont]
		
	mkTotBWScale $w.bw $name
	pack $w.addr $w.bw -side top -expand 1 -fill x
	
	set btns [frame $w.buttons]
	set config(pathnames,buttons,$name) $btns
	frame $btns.enc -highlightthickness 1 -borderwidth 2
	set config(pathnames,encoders,$name) \
		[menubutton $btns.enc.mb -menu $btns.enc.mb.m \
			-text "Format..." -font $f]
	mkEncoders $name $btns.enc.mb.m
	pack $btns.enc.mb

	frame $btns.glue
	pack $btns.enc -side left -anchor w
	checkbutton $btns.switching -text "Switching" -font $f \
		-highlightthickness 1 \
		-variable config($name,switching) -onvalue "on" -offvalue "off"
	pack $btns.glue -side left -expand 1 -fill x -anchor w
	pack $btns.switching -side right
	pack $btns -side top -expand 1 -fill x -padx 4 -pady 4
	if { $config($name,RC) == 0 || $config(vgw_engine) == ""} then {
		$btns.enc.mb configure -state disabled
		$btns.switching configure -state disabled
	}
}

proc mkSrcFrame w {
	frame $w -highlightthickness 2 -borderwidth 2
}	    

